/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"
#include "gl_mtex.h"
#include "gl_mirror.h"

int smoketexture;
int bloodtexture;

cvar_t gl_vertexlight = {"gl_vertexlight", "0", true};
cvar_t gl_aliasglow = {"gl_aliasglow", "1", true};
cvar_t gl_aliastrans = {"gl_aliastrans", "1", true};
cvar_t gl_fullbright = {"gl_fullbright", "1", true};
cvar_t gl_fullbright_alias = {"gl_fullbright_alias", "1", true};
cvar_t gl_fastbutugly = {"gl_fastbutugly", "0", true};
cvar_t r_detailsky = {"r_detailsky", "1", true};
cvar_t _r_skybox = {"_r_skybox", "none", true};
cvar_t _r_persist = {"_r_persist", "1", true};

// fog control
cvar_t gl_fogenable = {"gl_fogenable", "0", true};	// *don't* confuse this with the gl_fogenabled int
cvar_t gl_fogstart = {"gl_fogstart", "1024", true};
cvar_t gl_fogend = {"gl_fogend", "2048", true};
cvar_t gl_fogred = {"gl_fogred", "0.3", true};
cvar_t gl_foggreen = {"gl_foggreen", "0.3", true};
cvar_t gl_fogblue = {"gl_fogblue", "0.3", true};

// switch multitexturing on and off
cvar_t gl_multitexture = {"gl_multitexture", "1", true};

// dynamic light speedups
cvar_t gl_dlradmult = {"gl_dlradmult", "1.0", true};

// water waves & other water stuff
cvar_t r_waterwave = {"r_waterwave", "0.0", true};
cvar_t r_watershade = {"r_watershade", "1.0", true};
cvar_t r_waterdlight = {"r_waterdlight", "0", true};

// stop errors in zerstorer cutscenes - i had originally removed this, cos mhq locks viewsize
// at 120 - however, this causes error messages.  it's back, but it does nothing...
cvar_t viewsize = {"viewsize", "0", true};

// smoking torches
cvar_t r_torchsmoke = {"r_torchsmoke", "1.0", true};

// rotating skyboxes
cvar_t sky_rotate_x = {"sky_rotate_x", "0.0", true};
cvar_t sky_rotate_y = {"sky_rotate_y", "0.0", true};
cvar_t sky_rotate_z = {"sky_rotate_z", "0.0", true};

// automatic water translucency!
cvar_t r_autowatertrans = {"r_autowatertrans", "0", true};

// alias model light lerping
cvar_t gl_aliaslightlerp = {"gl_aliaslightlerp", "1", true};

// sky fogging
cvar_t gl_skyfog = {"gl_skyfog", "1", true};


// sin and cos tables from 0 to 1 in 0.0625 increments to speed up glow rendering
float glowcos[17] =
{
    1.000000,
    0.923879,
    0.707106,
    0.382682,
    -0.000002,
    -0.382686,
    -0.707109,
    -0.923881,
    -1.000000,
    -0.923878,
    -0.707103,
    -0.382678,
    0.000006,
    0.382689,
    0.707112,
    0.923882,
    1.000000,
};


float glowsin[17] =
{
    0.000000,
    0.382684,
    0.707107,
    0.923880,
    1.000000,
    0.923879,
    0.707105,
    0.382680,
    -0.000004,
    -0.382687,
    -0.707110,
    -0.923882,
    -1.000000,
    -0.923877,
    -0.707102,
    -0.382677,
    0.000008,
};


/*
==================
R_InitTextures
==================
*/
void	R_InitTextures (void)
{
    int		x,y, m;
    byte	*dest;

    // create a simple checkerboard texture for the default
    r_notexture_mip = Hunk_AllocName (sizeof(gltexture_t) + 16*16+8*8+4*4+2*2, "notexture");

    r_notexture_mip->width = r_notexture_mip->height = 16;
    r_notexture_mip->offsets[0] = sizeof(gltexture_t);
    r_notexture_mip->offsets[1] = r_notexture_mip->offsets[0] + 16*16;
    r_notexture_mip->offsets[2] = r_notexture_mip->offsets[1] + 8*8;
    r_notexture_mip->offsets[3] = r_notexture_mip->offsets[2] + 4*4;

    for (m=0 ; m<4 ; m++)
    {
        dest = (byte *)r_notexture_mip + r_notexture_mip->offsets[m];

        for (y=0 ; y< (16>>m) ; y++)
            for (x=0 ; x< (16>>m) ; x++)
            {
                if (  (y< (8>>m) ) ^ (x< (8>>m) ) )
                    *dest++ = 0;
                else
                    *dest++ = 0xff;
            }
    }
}

/*
==================
Initialize Particle Textures

MFAH - Includes new Smoke and Blood Textures
==================
*/

extern byte dottexture[16][16];
extern byte smoke_texture[32][32];
extern byte blood_texture[32][32];

void R_InitParticleTexture (void)
{
    int		x,y;
    byte	data[16][16][4];
    byte	smokedata[32][32][4];

    // hack to allow us to use pointers for gl_loadtexture
    int texhack = 64;

    // dot texture
    for (x = 0; x < 16; x++)
    {
        for (y = 0; y < 16; y++)
        {
            data[y][x][0] = 255;
            data[y][x][1] = 255;
            data[y][x][2] = 255;

            // i'm lazy and couldn't really be bothered rebuilding the texture...
            data[y][x][3] = (dottexture[x][y] * 114) / 100;
        }
    }

    particletexture = texture_extension_number++;

    GL_Bind (particletexture);

    glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 16, 16, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    // smoke texture
    for (x=0 ; x<32 ; x++)
    {
        for (y=0 ; y<32 ; y++)
        {
            smokedata[y][x][0] = 255;
            smokedata[y][x][1] = 255;
            smokedata[y][x][2] = 255;
            smokedata[y][x][3] = smoke_texture[x][y];
        }
    }

    // try to find a PCX or TGA first - if not present, load the default texture
    smoketexture = GL_LoadTexture ("smoketexture", &texhack, &texhack, NULL, false, true);

    if (smoketexture == -1)
    {
        smoketexture = texture_extension_number++;
        GL_Bind (smoketexture);

        glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, smokedata);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }

    // blood texture
    for (x=0 ; x<32 ; x++)
    {
        for (y=0 ; y<32 ; y++)
        {
            smokedata[y][x][0] = 255;
            smokedata[y][x][1] = 255;
            smokedata[y][x][2] = 255;
            smokedata[y][x][3] = blood_texture[x][y];
        }
    }

    // try to find a PCX or TGA first - if not present, load the default texture
    bloodtexture = GL_LoadTexture ("bloodtexture", &texhack, &texhack, NULL, false, true);

    if (bloodtexture == -1)
    {
        bloodtexture = texture_extension_number++;
        GL_Bind (bloodtexture);

        glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, smokedata);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
}

/*
===============
R_Init
===============
*/
void R_Init (void)
{
    extern byte *hunk_base;
    extern cvar_t gl_finish;
    byte data[16*16];
    int i;	// dummy

    // hack to allow us to use pointers for gl_loadtexture;
    int texhack = 16;

    Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);
    Cmd_AddCommand ("pointfile", R_ReadPointFile_f);

    Cvar_RegisterVariable (&r_norefresh);
    Cvar_RegisterVariable (&r_lightmap);
    Cvar_RegisterVariable (&r_drawentities);
    Cvar_RegisterVariable (&r_drawviewmodel);
    Cvar_RegisterVariable (&r_shadows);
    Cvar_RegisterVariable (&r_mirroralpha);
    Cvar_RegisterVariable (&r_wateralpha);
    Cvar_RegisterVariable (&r_dynamic);
    Cvar_RegisterVariable (&r_novis);
    Cvar_RegisterVariable (&r_speeds);

    Cvar_RegisterVariable (&gl_finish);
    Cvar_RegisterVariable (&gl_clear);
    Cvar_RegisterVariable (&gl_aliasglow);
    Cvar_RegisterVariable (&gl_aliastrans);
    Cvar_RegisterVariable (&gl_fullbright);
    Cvar_RegisterVariable (&gl_fullbright_alias);
    Cvar_RegisterVariable (&gl_fastbutugly);

    Cvar_RegisterVariable (&gl_cull);
    Cvar_RegisterVariable (&gl_smoothmodels);
    Cvar_RegisterVariable (&gl_affinemodels);
    Cvar_RegisterVariable (&gl_polyblend);
    Cvar_RegisterVariable (&gl_playermip);
    Cvar_RegisterVariable (&gl_nocolors);

    Cvar_RegisterVariable (&gl_keeptjunctions);
    Cvar_RegisterVariable (&gl_reporttjunctions);

    Cvar_RegisterVariable (&gl_doubleeyes);
    Cvar_RegisterVariable (&r_detailsky);
    Cvar_RegisterVariable (&_r_skybox);
    Cvar_RegisterVariable (&_r_persist);

    Cvar_RegisterVariable (&gl_vertexlight);

    // zerstorer cutscene error fix
    Cvar_RegisterVariable (&viewsize);

    // smoking torches
    Cvar_RegisterVariable (&r_torchsmoke);

    // fenix@io.com: register new cvars for model interpolation
    Cvar_RegisterVariable (&r_interpolate_alias);
    Cvar_RegisterVariable (&r_interpolate_brush);

    Cvar_RegisterVariable (&gl_fogenable);
    Cvar_RegisterVariable (&gl_fogstart);
    Cvar_RegisterVariable (&gl_fogend);
    Cvar_RegisterVariable (&gl_fogred);
    Cvar_RegisterVariable (&gl_foggreen);
    Cvar_RegisterVariable (&gl_fogblue);

    Cvar_RegisterVariable (&gl_multitexture);

    Cvar_RegisterVariable (&gl_dlradmult);

    Cvar_RegisterVariable (&r_waterwave);
    Cvar_RegisterVariable (&r_watershade);
    Cvar_RegisterVariable (&r_waterdlight);

    Cvar_RegisterVariable (&sky_rotate_x);
    Cvar_RegisterVariable (&sky_rotate_y);
    Cvar_RegisterVariable (&sky_rotate_z);

    Cvar_RegisterVariable (&r_autowatertrans);
    Cvar_RegisterVariable (&gl_aliaslightlerp);

    Cvar_RegisterVariable (&r_waterwarp);

    Cvar_RegisterVariable (&gl_skyfog);

    R_InitParticles ();
    R_InitParticleTexture ();

    memset (data, 108, sizeof (data));

    for (i = 0; i < (16 * 16); i += 2)
        data[i] = 74;

    notexture_num = i = GL_LoadTexture ("notexture", &texhack, &texhack, (byte *)data, true, false);

    r_notexture_mip->gl_texturenum = i;
    r_notexture_mip->fullbright = -1;

    // does this make assumptions about the texture extension number here?
    playertextures = texture_extension_number;
    texture_extension_number += 16;
}

/*
===============
R_TranslatePlayerSkin

Translates a skin texture by the per-player color lookup
===============
*/
// ignore translation cos it causes the psychedelic player head bug
// #define NO_TRANSLATE

// faster translation of player skins - does in place rather than sending through gl_upload8
#ifndef NO_TRANSLATE
#define FAST_TRANSLATE

#ifndef FAST_TRANSLATE
int GL_Upload8 (byte *data, int width, int height,  qboolean mipmap, qboolean alpha);
#endif

extern byte *player_texels;

void R_TranslatePlayerSkin (int playernum)
{
    int         top, bottom;
    byte        *original, translate[256];
    int         i, s;
    model_t	    *model;
    aliashdr_t  *paliashdr;
    int         inwidth, inheight;

#ifdef FAST_TRANSLATE
    int         j;
    unsigned    scaled_width, scaled_height;
    unsigned    frac;
    unsigned    fracstep;
    unsigned    *out;
    unsigned    pixels[512*256];
    unsigned    translate32[256];
    byte        *inrow;
#else
    byte        translated[320 * 200];
#endif

    top = cl.scores[playernum].colors & 0xf0;
    bottom = (cl.scores[playernum].colors & 15) << 4;

    for (i = 0; i < 256; i++) translate[i] = i;

    for (i = 0; i < 16; i++)
    {
        if (top < 128) translate[TOP_RANGE + i] = top + i;
        else translate[TOP_RANGE + i] = top + 15 - i;

        if (bottom < 128) translate[BOTTOM_RANGE + i] = bottom + i;
        else translate[BOTTOM_RANGE + i] = bottom + 15 - i;
    }

    // locate the original skin pixels
    currententity = &cl_entities[1 + playernum];
    model = currententity->model;

    if (!model) return;
    if (model->type != mod_alias) return;

    paliashdr = (aliashdr_t *)Mod_Extradata (model);
    s = paliashdr->skinwidth * paliashdr->skinheight;

#ifdef FAST_TRANSLATE
	if (currententity->skinnum < 0 || currententity->skinnum >= paliashdr->numskins)
	{
		Con_DPrintf ("(%d): Invalid player skin #%d\n", playernum, currententity->skinnum);
		original = (byte *)paliashdr + paliashdr->texels[0];
	}
	else
	{
		original = (byte *)paliashdr + paliashdr->texels[currententity->skinnum];
	}
#endif

    if (s & 3)
    {
        // exit gracefully
        Con_Printf ("s & 3\n");
        return;
    }

    inwidth = paliashdr->skinwidth;
    inheight = paliashdr->skinheight;

    // no need to bother checking oldtexture here
    GL_Bind (playertextures + playernum + 1);

#ifdef FAST_TRANSLATE
    // because this happens during gameplay, do it fast
    // instead of sending it through gl_upload 8

    // this always resamples player skins up to either 512 x 256, or the max texture
    // size if that's smaller.
    scaled_width = gl_max_size.value < 512 ? gl_max_size.value : 512;
    scaled_height = gl_max_size.value < 256 ? gl_max_size.value : 256;

    // allow users to crunch sizes down even more if they want
    scaled_width >>= (int)gl_playermip.value;
    scaled_height >>= (int)gl_playermip.value;

    for (i = 0; i < 256; i++) translate32[i] = d_8to24table[translate[i]];

    out = pixels;
    fracstep = inwidth * 0x10000 / scaled_width;

    for (i = 0; (unsigned) i < scaled_height; i++, out += scaled_width)
    {
        inrow = original + inwidth * (i * inheight / scaled_height);
        frac = fracstep >> 1;

        for (j = 0; (unsigned) j < scaled_width; j += 4)
        {
            out[j] = translate32[inrow[frac >> 16]];
            frac += fracstep;
            out[j+1] = translate32[inrow[frac >> 16]];
            frac += fracstep;
            out[j+2] = translate32[inrow[frac >> 16]];
            frac += fracstep;
            out[j+3] = translate32[inrow[frac >> 16]];
            frac += fracstep;
        }
    }

    glTexImage2D (GL_TEXTURE_2D, 0, gl_solid_format, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
#else
    for (i = 0; i < s; i += 4)
    {
        translated[i] = translate[player_texels[i]];
        translated[i+1] = translate[player_texels[i+1]];
        translated[i+2] = translate[player_texels[i+2]];
        translated[i+3] = translate[player_texels[i+3]];
    }

    // don't mipmap these, because it takes too long
    i = GL_Upload8 (translated, paliashdr->skinwidth, paliashdr->skinheight, false, false);
#endif
}
#else
void R_TranslatePlayerSkin (int playernum)
{
    return;
}
#endif


/*
===============
R_NewMap
===============
*/
extern qboolean r_mapshot;

void R_NewMap (void)
{
    int		i;

    for (i=0 ; i<256 ; i++)
        d_lightstylevalue[i] = 264;		// normal light value

    rotate_x = 0;
    rotate_y = 0;
    rotate_z = 0;

    memset (&r_worldentity, 0, sizeof(r_worldentity));
    r_worldentity.model = cl.worldmodel;

    // clear out efrags in case the level hasn't been reloaded
    // FIXME: is this one short?
    for (i=0 ; i<cl.worldmodel->numleafs ; i++)
        cl.worldmodel->leafs[i].efrags = NULL;

    r_viewleaf = NULL;
    R_ClearParticles ();

    GL_BuildLightmaps ();

    // identify sky texture
    skytexturenum = -1;

    for (i=0 ; i<cl.worldmodel->numtextures ; i++)
    {
        if (!cl.worldmodel->textures[i]) continue;
        if (!strncmp(cl.worldmodel->textures[i]->name,"sky",3)) skytexturenum = i;
        cl.worldmodel->textures[i]->texturechain = NULL;
    }

    // map configuration scripts have been removed cos the potential for abuse is just so
    // enormous.

    // reset r_wateralpha in case the new map hasn't been properly vised.  MHScript and auto
    // water translucency will change this later on anyway...
    Cvar_SetValue ("r_wateralpha", 1.0);

#ifdef USE_SCRIPTS
    // automatically run mhscripts
    Cbuf_AddText ("script map false");
    Cbuf_AddText ("script mdl false");
    Cbuf_AddText ("script tex false");
    Cbuf_AddText ("script hud false");
#endif

    SCR_CenterPrint (cl.levelname);

    r_mapshot = false;
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
    int			i;
    float		start, stop, time;

    glDrawBuffer  (GL_FRONT);
    glFinish ();

    start = Sys_FloatTime ();

    for (i=0 ; i<128 ; i++)
    {
        r_refdef.viewangles[1] = i/128.0*360.0;
        R_RenderView ();
    }

    glFinish ();
    stop = Sys_FloatTime ();
    time = stop-start;
    Con_Printf ("%f seconds (%f fps)\n", time, 128/time);

    glDrawBuffer  (GL_BACK);
    GL_EndRendering ();
}


void D_FlushCaches (void)
{
}


