/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// r_surf.c: surface-related refresh code - the lightmap stuff has been moved to gl_lightmap.c
// colossal chunks of this file have been rewritten - it's barely recognisable now.  i've
// cleaned up an awful lot and made it much more efficient.

#include "quakedef.h"
#include "gl_mtex.h"
#include "gl_lightmap.h"
#include "gl_mirror.h"

glmsurface_t *skychain = NULL;
glmsurface_t *waterchain = NULL;
glmsurface_t *slimechain = NULL;
glmsurface_t *lavachain = NULL;
glmsurface_t *telechain = NULL;
glmsurface_t *windowchain = NULL;
glmsurface_t *fogchain = NULL;
glmsurface_t *pervchain = NULL;

int oldtexture;
int oldtexnum_mtex2;
int	skytexturenum;
int gl_useclight;

// make it easier to find out if we're drawing multitextured
qboolean MTexUse;

// function pointers for surface drawing
void (*ChainDrawer) ();
void (*BMPolyDrawer) ();

qboolean force_vis = false;
qboolean water_present = false;
qboolean evil_hack;

/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
MH - fixed the mis-spelling of 'relative' in here!!!
===============
*/
gltexture_t *R_TextureAnimation (gltexture_t *base)
{
    int		    relative;
    int		    count;
    gltexture_t *next;

    if (currententity->frame)
    {
        if (base->alternate_anims)
        {
            base = base->alternate_anims;
        }
    }

    // this has to be done second because there could be only one alternate animation!!!!!
    if (!base->anim_total)
    {
        return base;
    }
    relative = (int)(cl.time*10) % base->anim_total;

    count = 0;

    while (base->anim_min > relative || base->anim_max <= relative)
    {
        next = base->anim_next;

        // mh: made these safe
        if (!next)
        {
            return base;				// broken cycle
        }

        if (++count > 100)
        {
            return base;		// 'infinite' cycle
        }
        base = next;
    }
    return base;
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/

lpMTexFUNC MTexCoord2f = NULL;
lpSelTexFUNC SelectTextureMTex = NULL;

void R_SetupLightmapDraw (void)
{
    glEnable (GL_BLEND);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
    glDepthMask (0);
    glBlendFunc (GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
}

void R_ShutdownLightmapDraw (void)
{
    glDisable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask (1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

void R_UploadLightmap (int lightmapnum)
{
    glRect_t *theRect;

    lightmap_modified[lightmapnum] = false;
    theRect = &lightmap_rectchange[lightmapnum];

    glTexSubImage2D (GL_TEXTURE_2D, 0, 0, theRect->t, BLOCK_WIDTH, theRect->h, gl_lightmap_format, GL_UNSIGNED_BYTE, lightmaps + (lightmapnum * BLOCK_HEIGHT + theRect->t) * BLOCK_WIDTH * lightmap_bytes);

    theRect->l = BLOCK_WIDTH;
    theRect->t = BLOCK_HEIGHT;
    theRect->h = 0;
    theRect->w = 0;
}

void R_DrawLightmapPolys (glmsurface_t *s)
{
    int         i;
    glpoly_t    *p;
    float       *v;

    i = s->lightmaptexturenum;

    // this one can potentially stay the same between calls to this function...
    if (oldtexture != lightmap_textures + i || lightmap_modified[i])
    {
        glBindTexture (GL_TEXTURE_2D, lightmap_textures + i);
        oldtexture = lightmap_textures + i;

        if (lightmap_modified[i])
        {
            R_UploadLightmap (i);
        }
    }
    glBegin (GL_POLYGON);

    p = s->polys;
    v = p->verts[0];

    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE)
    {
        glTexCoord2fv (&v[5]);
        glVertex3fv (v);
    }
    glEnd ();
}

void R_DrawFullBrightPolys (glmsurface_t *s)
{
    glpoly_t	*p;
    float		*v;
    int			i;

    p = s->polys;
    v = p->verts[0];

    glBegin (GL_POLYGON);

    for (i=0 ; i<p->numverts ; i++, v+= VERTEXSIZE)
    {
        glTexCoord2fv (&v[3]);
        glVertex3fv (v);
    }
    glEnd ();
}

void R_DrawVertexLightPolys (glmsurface_t *s)
{
    glpoly_t	*p;
    float		*v;
    int			i;

    glBegin (GL_POLYGON);

    p = s->polys;
    v = p->verts[0];

    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE)
    {
        glTexCoord2fv (&v[3]);
        glColor3fv (&v[8]);
        glVertex3fv (v);
    }
    glEnd ();
}

void R_DrawMTexPolys (glmsurface_t *s)
{
    glpoly_t	*p;
    float		*v;
    int			i;

    i = s->lightmaptexturenum;

    // this time round, we will use oldtexnum_mtex2 cos the lightmap texture used for
    // any given surface is a subimage of a larger texture - which may not necessarily
    // change... it will also need to change if the lightmap becomes modified.
    if (lightmap_textures + i != oldtexnum_mtex2 || lightmap_modified[i])
    {
        glBindTexture (GL_TEXTURE_2D, lightmap_textures + i);
        oldtexnum_mtex2 = lightmap_textures + i;

        if (lightmap_modified[i])
        {
            R_UploadLightmap (i);
        }
    }
    glBegin(GL_POLYGON);

    p = s->polys;
    v = p->verts[0];

    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE)
    {
        MTexCoord2f (GL_MTexture0, v[3], v[4]);
        MTexCoord2f (GL_MTexture1, v[5], v[6]);
        glVertex3fv (v);
    }
    glEnd ();
}


// fullbright colours
// ==================
void R_SetupFullBrightColourDraw (void)
{
    // shut down multitexturing
    if (MTexUse)
    {
        glDisable(GL_TEXTURE_2D);
        SelectTextureMTex (GL_MTexture0);
    }

    // shut down fog
    if (gl_fogenabled)
    {
        glDisable (GL_FOG);
    }
    glEnable (GL_BLEND);
}


void R_ShutdownFullBrightColourDraw (void)
{
    glDisable (GL_BLEND);

    // start up multitexturing
    if (MTexUse)
    {
        SelectTextureMTex (GL_MTexture1);
        glEnable(GL_TEXTURE_2D);
    }

    // start up fog
    if (gl_fogenabled)
    {
        glEnable (GL_FOG);
    }
}

void DrawFullBrightColourChain (glmsurface_t *s, gltexture_t *t)
{
    GL_Bind (t->fullbright);

    for (/**/; s; s = s->texturechain)
    {
        R_DrawFullBrightPolys (s);
    }
}

/*
================
DrawTextureChains
================
*/
void DrawFullBrightChain (glmsurface_t *s, gltexture_t *t)
{
    GL_Bind (t->gl_texturenum);

    for (/**/; s; s = s->texturechain)
    {
        R_DrawFullBrightPolys (s);
    }
}

void DrawVertexLightChain (glmsurface_t *s, gltexture_t *t)
{
    GL_Bind (t->gl_texturenum);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    for (/**/; s; s = s->texturechain)
    {
        R_DrawVertexLightPolys (s);
    }
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glColor3f (1, 1, 1);
}

void DrawMTexChain (glmsurface_t *s, gltexture_t *t)
{
    SelectTextureMTex (GL_MTexture0);
    GL_Bind (t->gl_texturenum);
    SelectTextureMTex (GL_MTexture1);

    for (/**/; s; s = s->texturechain)
    {
        R_DrawMTexPolys (s);
    }
}

void DrawNoMTexChain (glmsurface_t *s, gltexture_t *t)
{
    GL_Bind (t->gl_texturenum);

    for (/**/; s; s = s->texturechain)
    {
        R_DrawFullBrightPolys (s);
    }
    R_SetupLightmapDraw ();

    for (/**/; s; s = s->texturechain)
    {
        R_DrawLightmapPolys (s);
    }
    R_ShutdownLightmapDraw ();
}


void DrawTextureChains (void)
{
    int             i;
    glmsurface_t    *s;
    gltexture_t     *t, *tx;

    for (i = 0; i < cl.worldmodel->numtextures; i++)
    {
        t = cl.worldmodel->textures[i];
        if (!t) continue;
        s = t->texturechain;
        if (!s) continue;

        tx = R_TextureAnimation (t);

        (*ChainDrawer) (s, tx);

        if (!gl_fullbright.value)
        {
            t->texturechain = NULL;
        }
    }
}

void DrawFullBrightTextureChains (void)
{
    int i;
    glmsurface_t *s;
    gltexture_t *t, *tx;

    if (!gl_fullbright.value) return;

    // draw fullbrights faster in the world
    R_SetupFullBrightColourDraw ();

    for (i = 0; i < cl.worldmodel->numtextures; i++)
    {
        t = cl.worldmodel->textures[i];
        if (!t) continue;
        s = t->texturechain;
        if (!s) continue;

        tx = R_TextureAnimation (t);

        if (tx->fullbright != -1)
        {
            DrawFullBrightColourChain (s, tx);
        }
        t->texturechain = NULL;
    }
    R_ShutdownFullBrightColourDraw ();
}

/*
=================
R_DrawBrushModel
=================
*/
int R_CheckSpecialBModelSurfs (glmsurface_t *s)
{
    if (s->flags & SURF_DRAWPERV)
    {
        s->texturechain = pervchain;
        pervchain = s;
        return 1;
    }

    // fixme: this invalidates moving liquid brush models (which you shouldn't really be
    // doing anyway - unfortunately the whole community is swamped with hacks and exploitations
    // of loopholes like this...)
    if (s->flags & SURF_DRAWTURB)
    {
        if (s->flags & SURF_LAVA)
        {
            s->texturechain = lavachain;
            lavachain = s;
        }
        else if (s->flags & SURF_SLIME)
        {
            water_present = true;
            s->texturechain = slimechain;
            slimechain = s;
        }
        else if (s->flags & SURF_TELE)
        {
            s->texturechain = telechain;
            telechain = s;
        }
        else
        {
            water_present = true;
            s->texturechain = waterchain;
            waterchain = s;
        }
        return 1;
    }

    if (s->flags & SURF_DRAWSKY)
    {
        // sorry - but it's the lesser of 2 evils...
        if (!gl_useskybox)
        {
            EmitBothSkyLayers (s);
        }
        return 1;
    }

    if (s->flags & SURF_WINDOW)
    {
        s->texturechain = windowchain;
        windowchain = s;
        return 1;
    }

    if (s->flags & SURF_FOG)
    {
        s->texturechain = fogchain;
        fogchain = s;
        return 1;
    }

    if (s->flags & SURF_TRANS)
    {
        if (MTexUse)
        {
            glDisable (GL_TEXTURE_2D);
            SelectTextureMTex (GL_MTexture0);
        }
        glEnable (GL_ALPHA_TEST);
        glAlphaFunc (GL_GEQUAL, 0.005);

        R_RenderDynamicVertexLightmaps (s);

        DrawVertexLightChain (s, R_TextureAnimation (s->texinfo->texture));

        glAlphaFunc (GL_GEQUAL, 0.666);
        glDisable (GL_ALPHA_TEST);

        if (MTexUse)
        {
            SelectTextureMTex (GL_MTexture1);
            glEnable (GL_TEXTURE_2D);
        }
        return 1;
    }
    return 0;
}

void R_DrawBrushModel (entity_t *e)
{
    int			    k;
    vec3_t		    mins, maxs;
    int			    i;
    glmsurface_t	*psurf;
    float		    dot;
    mplane_t	    *pplane;
    model_t		    *clmodel;
    qboolean	    rotated;
    gltexture_t     *tx;

    currententity = e;
    clmodel = e->model;

    if (e->angles[0] || e->angles[1] || e->angles[2])
    {
        rotated = true;
        for (i=0 ; i<3 ; i++)
        {
            mins[i] = e->origin[i] - clmodel->radius;
            maxs[i] = e->origin[i] + clmodel->radius;
        }
    }
    else
    {
        rotated = false;
        VectorAdd (e->origin, clmodel->mins, mins);
        VectorAdd (e->origin, clmodel->maxs, maxs);
    }

    if (R_CullBox (mins, maxs)) return;

    glColor3f (1,1,1);

    VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

    if (rotated)
    {
        vec3_t	temp;
        vec3_t	forward, right, up;

        VectorCopy (modelorg, temp);
        AngleVectors (e->angles, forward, right, up);
        modelorg[0] = DotProduct (temp, forward);
        modelorg[1] = -DotProduct (temp, right);
        modelorg[2] = DotProduct (temp, up);
    }

    psurf = &clmodel->surfaces[clmodel->firstmodelsurface];

    // calculate dynamic lighting for bmodel if it's not an
    // instanced model
    if (clmodel->firstmodelsurface != 0)
    {
        for (k=0 ; k<MAX_DLIGHTS ; k++)
        {
            if (!r_dynamic.value) break;

            if ((cl_dlights[k].die < cl.time) || (!cl_dlights[k].radius)) continue;

            if (cl_dlights[k].type == DL_BLEND) continue;

            R_MarkLights (&cl_dlights[k], 1<<k, clmodel->nodes + clmodel->hulls[0].firstclipnode);
        }
    }

    glPushMatrix ();
    e->angles[0] = -e->angles[0];	// stupid quake bug

    // rotating brush model interpolation!!!!
    if (r_interpolate_brush.value)
        R_BlendedRotateForEntity (e, 0);
    else R_RotateForEntity (e);

    e->angles[0] = -e->angles[0];	// stupid quake bug

    // to do - go through the surfs, sorting them into texturechains, then go through the
    // texturechains for drawing - problem is, there's usually so few textures in any given
    // brush model that the speedup would be negligible (it might even be slower)
    for (i=0 ; i<clmodel->nummodelsurfaces ; i++, psurf++)
    {
        // find which side of the node we are on
        pplane = psurf->plane;
        dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

        // draw the polygon
        if (((psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(psurf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
        {
            // to do - if it's a liquid texture, build a new waterchain or whatever, and
            // draw it now. (in case we have moving brushmodels with liquid textures)
            if (R_CheckSpecialBModelSurfs (psurf)) continue;

            tx = R_TextureAnimation (psurf->texinfo->texture);
            R_RenderDynamicLightmaps (psurf);

            // set up a dummy texturechain in order to fool chaindrawer into thinking
            // we have an actual texturechain.  this lets me use the same functions
            // for texturechains and for single textures.
            psurf->texturechain = psurf->texinfo->texture->texturechain;
            psurf->texinfo->texture->texturechain = psurf;

            (*ChainDrawer) (psurf, tx);

            // fullbrights are now drawn seperately from the main texture chain, so we have to
            // hack them in here... this is much slower than for the world, but there's not
            // much we can do about it
            if (tx->fullbright != -1 && gl_fullbright.value)
            {
                R_SetupFullBrightColourDraw ();
                DrawFullBrightColourChain (psurf, tx);
                R_ShutdownFullBrightColourDraw ();
            }

            // shut down the texturechain to stop it flooding memory - both of these probably
            // aren't necessary, but it does no harm... (well, maybe a teensy bit of speed)
            psurf->texinfo->texture->texturechain = NULL;
            psurf->texturechain = NULL;
        }
    }
    glPopMatrix ();
}


/*
================
R_DrawTranslucentSurfaces
Water & Windows - Water, whether translucent or not, is *always* drawn at this stage!
================
*/

extern cvar_t r_waterwave;
extern cvar_t r_watershade;
extern cvar_t r_waterdlight;

void SetTranslucentLiquid (float liquid_alpha, float liquid_brightness)
{
    if (liquid_alpha >= 1.0 || hlbsp)
    {
        // halflife does translucent water differently, so just leave it
        // fully opaque for now.
        glColor3f (liquid_brightness, liquid_brightness, liquid_brightness);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }
    else
    {
        glEnable (GL_BLEND);
        glDepthMask (0);
        glColor4f (liquid_brightness, liquid_brightness, liquid_brightness, liquid_alpha);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }
}

void ResetTranslucentLiquid (float liquid_alpha)
{
    if (liquid_alpha >= 1.0 || hlbsp)
    {
        // halflife does translucent water differently, so just leave it
        // fully opaque for now.
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        glColor3f (1, 1, 1);
    }
    else
    {
        glDepthMask (1);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        glColor4f (1, 1, 1, 1);
        glDisable (GL_BLEND);
    }
}

float r_waterfog_colour[4];
float r_waterfog_density;

#define NO_TRANS	1.0
#define NO_WAVE		0.0
#define NO_SHADE	0.8

void R_DrawLiquidChain (glmsurface_t *liquid_chain, float temp_wave, float vlight_base, float trans_lev, qboolean save_fog)
{
    glmsurface_t	*s;
    glpoly_t		*p;
    int				i;
    float			*v;

    if (!liquid_chain) return;

    // evil hack for underwater water surfaces with auto trans water
    if (evil_hack)
        SetTranslucentLiquid (NO_TRANS, vlight_base);
    else
        SetTranslucentLiquid (trans_lev, vlight_base);

    for (s = liquid_chain; s; s = s->texturechain)
    {
        if (s->texinfo->texture->gl_texturenum != oldtexture)
        {
            glBindTexture (GL_TEXTURE_2D, s->texinfo->texture->gl_texturenum);
            oldtexture = s->texinfo->texture->gl_texturenum;
        }

        // initialize vertex lighting for the surface
        for (p = s->warp_polys; p; p = p->next)
		{
            for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
			{
                v[8] = v[9] = v[10] = vlight_base;
            }
        }

        // hmm does no difference
        // but disabling the above fucks up watersurfs.
        if (r_waterdlight.value)
        {
            R_RenderDynamicVertexLightmaps (s);
        }

        // then draw it
        EmitWaterPolys (s, temp_wave, trans_lev, r_waterwarp.value);

        // no underwater fogging on teleport textures
        if (save_fog)
        {
            // save these out for underwater fogging... this can potentially fuck up if
            // there are 2 or more different types of liquid being drawn in the same pvs.
            // mhquake 6 - not any more, matey!
            r_waterfog_colour[0] = s->texinfo->texture->water_fog_colour[0];
            r_waterfog_colour[1] = s->texinfo->texture->water_fog_colour[1];
            r_waterfog_colour[2] = s->texinfo->texture->water_fog_colour[2];
            r_waterfog_colour[3] = s->texinfo->texture->water_fog_density;	// oh well...
            r_waterfog_density = s->texinfo->texture->water_fog_density;
        }
    }

    // evil hack for underwater water surfaces with auto trans water
    if (evil_hack)
        ResetTranslucentLiquid (NO_TRANS);
    else
        ResetTranslucentLiquid (trans_lev);
}

void R_SetupOtherTransBlend (float blend_alpha)
{
    glEnable (GL_BLEND);
    glDepthMask (0);
    glColor4f (1, 1, 1, blend_alpha);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void R_ShutdownOtherTransBlend (void)
{
    glDepthMask (1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glColor4f (1, 1, 1, 1);
    glDisable (GL_BLEND);
}

void R_DrawWindows (void)
{
    if (windowchain)
    {
        glmsurface_t *s;

        R_SetupOtherTransBlend (0.4);

        for (s = windowchain; s; s = s->texturechain)
        {
            if (s->texinfo->texture->gl_texturenum != oldtexture)
            {
                glBindTexture (GL_TEXTURE_2D, s->texinfo->texture->gl_texturenum);
                oldtexture = s->texinfo->texture->gl_texturenum;
            }
            R_DrawFullBrightPolys (s);
        }
        windowchain = NULL;

        R_ShutdownOtherTransBlend ();
    }
}


void R_DrawPerv (void)
{
    if (pervchain)
    {
        int			i;
        glmsurface_t *s;
        glpoly_t	*p;
        float		*v;
        float       nv[2];

        R_SetupOtherTransBlend (0.8);

        for (s = pervchain; s; s = s->texturechain)
        {
            if (s->texinfo->texture->gl_texturenum != oldtexture)
            {
                glBindTexture (GL_TEXTURE_2D, s->texinfo->texture->gl_texturenum);
                oldtexture = s->texinfo->texture->gl_texturenum;
            }

            for (p = s->warp_polys; p; p = p->next)
            {
                glBegin (GL_POLYGON);

                for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
                {
                    nv[0] = v[3];
                    nv[1] = v[4];

                    Calc_PervModys (&nv[0], &nv[1], 40.75);

                    glTexCoord2fv (&nv[0]);
                    glVertex3fv (v);
                }
                glEnd ();
            }
        }
        pervchain = NULL;

        R_ShutdownOtherTransBlend ();
    }
}

void R_DrawFog (void)
{
    // doesn't really work right now
    if (fogchain)
    {
        glmsurface_t *s;

        R_SetupOtherTransBlend (0.2);

        for (s = fogchain; s; s = s->texturechain)
        {
            if (s->texinfo->texture->gl_texturenum != oldtexture)
            {
                glBindTexture (GL_TEXTURE_2D, s->texinfo->texture->gl_texturenum);
                oldtexture = s->texinfo->texture->gl_texturenum;
            }
            R_DrawFullBrightPolys (s);
        }
        fogchain = NULL;

        R_ShutdownOtherTransBlend ();
    }
}

void R_DrawTranslucentSurfaces (void)
{
    // bound r_waterwave to sensible values
    if (r_waterwave.value > 3.0)
        Cvar_SetValue ("r_waterwave", 3);
    else if (r_waterwave.value < 0.0)
        Cvar_SetValue ("r_waterwave", 0);

    // bound r_watershade to sensible values
    if (r_watershade.value > 0.8)
        Cvar_SetValue ("r_watershade", 0.8);
    else if (r_watershade.value < 0.0)
        Cvar_SetValue ("r_watershade", 0);

    // calc the warp sin tables
	Calc_WarpModys ();

    // go back to the world matrix
    glLoadMatrixf (r_world_matrix);

    R_DrawLiquidChain (telechain, NO_WAVE, NO_SHADE, NO_TRANS, false);
    R_DrawLiquidChain (waterchain, r_waterwave.value, r_watershade.value, r_wateralpha.value, inwater);
    R_DrawLiquidChain (lavachain, r_waterwave.value, NO_SHADE, NO_TRANS, inwater);
    R_DrawLiquidChain (slimechain, r_waterwave.value, NO_SHADE, r_wateralpha.value, inwater);

    // can't reset these to NULL in R_DrawLiquidChain cos the engine becomes very unhappy...
    telechain = NULL;
    waterchain = NULL;
    lavachain = NULL;
    slimechain = NULL;

    // other translucent surfaces are also drawn here
    R_DrawWindows ();
    R_DrawPerv ();
    R_DrawFog ();
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/


/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
    int			c, side;
    mplane_t	*plane;
    glmsurface_t	*surf, **mark;
    mleaf_t		*pleaf;
    double		dot;

    if (node->contents == CONTENTS_SOLID)
        return;		// solid

    if (node->visframe != r_visframecount)
        return;

    if (R_CullBox (node->minmaxs, node->minmaxs+3))
        return;

    // if a leaf node, draw stuff
    if (node->contents < 0)
    {
        pleaf = (mleaf_t *)node;

        mark = pleaf->firstmarksurface;
        c = pleaf->nummarksurfaces;

        if (c)
        {
            do
            {
                (*mark)->visframe = r_framecount;
                mark++;
            }
            while (--c);
        }

        // deal with model fragments in this leaf
        if (pleaf->efrags)
            R_StoreEfrags (&pleaf->efrags);

        return;
    }

    // node is just a decision point, so go down the apropriate sides
    plane = node->plane;

    switch (plane->type)
    {
    case PLANE_X:
        dot = modelorg[0] - plane->dist;
        break;
    case PLANE_Y:
        dot = modelorg[1] - plane->dist;
        break;
    case PLANE_Z:
        dot = modelorg[2] - plane->dist;
        break;
    default:
        dot = DotProduct (modelorg, plane->normal) - plane->dist;
        break;
    }

    if (dot >= 0)
        side = 0;
    else
        side = 1;

    // recurse down the children, front side first
    R_RecursiveWorldNode (node->children[side]);

    // draw stuff
    c = node->numsurfaces;

    if (c)
    {
        surf = cl.worldmodel->surfaces + node->firstsurface;

        if (dot < 0 -BACKFACE_EPSILON)
            side = SURF_PLANEBACK;
        else if (dot > BACKFACE_EPSILON)
            side = 0;

        for (; c; c--, surf++)
        {
            if (surf->visframe != r_framecount)
                continue;

            if (!(surf->flags & SURF_UNDERWATER) && ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)))
                continue;		// wrong side

            // we don't check windows and fog cos they're only in brush models...
            if (surf->flags & SURF_DRAWSKY)
            {
                surf->texturechain = skychain;
                skychain = surf;
                continue;
            }
            else if (surf->flags & SURF_DRAWTURB)
            {
                if (surf->flags & SURF_LAVA)
                {
                    surf->texturechain = lavachain;
                    lavachain = surf;
                }
                else if (surf->flags & SURF_TELE)
                {
                    surf->texturechain = telechain;
                    telechain = surf;
                }
                else if (surf->flags & SURF_SLIME)
                {
                    water_present = true;
                    surf->texturechain = slimechain;
                    slimechain = surf;
                }
                else
                {
                    water_present = true;
                    surf->texturechain = waterchain;
                    waterchain = surf;
                }

                continue;
            }
#ifdef MH_MIRRORS
            else if (r_mirroralpha.value < 1.0 && !mirror_render && surf->flags & SURF_MIRROR)
            {
                mirror = true;
                surf->texturechain = mirrorchain;
                mirrorchain = surf;
                continue;
            }
#endif
            else
            {
                surf->texturechain = surf->texinfo->texture->texturechain;
                surf->texinfo->texture->texturechain = surf;
            }

            // check for modified lightmaps on this surface - i get about 3 extra fps by
            // doing this here...
            R_RenderDynamicLightmaps (surf);
        }
    }

    // recurse down the back side
    R_RecursiveWorldNode (node->children[!side]);
}

// fixme: these names are misleading...
void R_ShutdownSkyFog (void)
{
    if (gl_fogenabled && !gl_skyfog.value)
	{
        glDisable (GL_FOG);
	}
    else if (gl_fogenabled && gl_skyfog.value)
    {
        glDisable (GL_FOG);
        glFogf (GL_FOG_START, gl_fogstart.value);
        glFogf (GL_FOG_END, gl_fogend.value * 1.25);
        glEnable (GL_FOG);
    }
}

void R_StartupSkyFog (void)
{
    if (gl_fogenabled && !gl_skyfog.value)
	{
        glEnable (GL_FOG);
	}
    else if (gl_fogenabled && gl_skyfog.value)
    {
        glDisable (GL_FOG);
        glFogf (GL_FOG_START, gl_fogstart.value);
        glFogf (GL_FOG_END, gl_fogend.value);
        glEnable (GL_FOG);
    }
}

void R_DrawWorld (void)
{
    entity_t	ent;

    oldtexture = -1;
    oldtexnum_mtex2 = -1;

    memset (&ent, 0, sizeof(ent));
    ent.model = cl.worldmodel;

    VectorCopy (r_refdef.vieworg, modelorg);

    currententity = &ent;
    glColor3f (1,1,1);

    // only done once per frame, and so trivial that it doesn't matter if it needs
    // to be done or not
    R_ClearSkyBox ();

    R_RecursiveWorldNode (cl.worldmodel->nodes);

    if (skychain && !gl_useskybox)
    {
        // don't fog the sky
        R_ShutdownSkyFog ();
        R_DrawSkyChain (skychain);
        R_StartupSkyFog ();
        skychain = NULL;
    }
#ifdef MH_MIRRORS
    else if (skychain && gl_useskybox && mirror)
    {
        R_DrawOldSkyboxChain (skychain);
    }
#endif

    R_SetupDraw (true);
    DrawTextureChains ();
    DrawFullBrightTextureChains ();

    if (gl_useskybox)
    {
        // have to do this so the skybox will draw properly
        R_ShutdownDraw ();

        if (mirror)
        {
            R_DrawOldSkyBox ();
        }
        else
        {
            R_DrawSkyBox ();
        }
        R_SetupDraw (true);
    }
}

/*
===============
R_MarkLeaves
===============
*/
void R_MarkLeaves (void)
{
    byte	*vis;
    mnode_t	*node;
    int		i;
    byte	solid[4096];

    // if we're in the same leaf as we were previously, don't update visibility
    // bug - if r_novis changes from 1 to 0 this will crap out.
    if (r_oldviewleaf == r_viewleaf && !r_novis.value)
    {
        if (!force_vis) return;
    }

    if (mirror)	return;

    r_visframecount++;
    r_oldviewleaf = r_viewleaf;

    evil_hack = false;

    if (r_novis.value)
    {
        vis = solid;
        memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
    }
    else
    {
        vis = Mod_LeafPVS (r_viewleaf, cl.worldmodel);
    }

    for (i = 0; i < cl.worldmodel->numleafs; i++)
    {
        // automatic water translucency
        if ((vis[i >> 3] & (1 << ( i & 7)))	|| (r_autowatertrans.value && water_present	&& (cl.worldmodel->leafs[i + 1].contents == CONTENTS_WATER || cl.worldmodel->leafs[i + 1].contents == CONTENTS_SLIME)))
        {
            // if we're currently underwater and the leaf is not normally visible,
            // (a) don't draw it (all water surfaces are opaque when underwater), and
            // (b) set r_wateralpha to 1.  this is a nasty hack but it prevents such
            // nonsense as underwater fog rendering incorrectly, and massive slowdowns
            // by drawing all the non-underwater leafs.
            if ((!(vis[i >> 3] & (1 << ( i & 7))) && (r_viewleaf->contents == CONTENTS_WATER || r_viewleaf->contents == CONTENTS_SLIME)) && !r_novis.value)
            {
                // the next out-of-water leaf will reset this to translucent.  r_wateralpha 1
                // doesn't work properly here, so we must use an evil hack instead.
                evil_hack = true;
                goto underwater_hack;
            }
            node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

            do
            {
                if (node->visframe == r_visframecount) break;
                node->visframe = r_visframecount;
                node = node->parent;
            }
            while (node);
        }
underwater_hack:;
    }
}

void R_SetupDraw (qboolean findfunc)
{
    if (findfunc)
    {
        // set the correct drawing functions - findfunc is always true in the current engine,
        // but has been left in just in case it's ever necessary to do this again
        if (gl_vertexlight.value)
        {
            MTexUse = false;
            ChainDrawer = DrawVertexLightChain;
        }
        else if (gl_mtexable && gl_multitexture.value)
        {
            MTexUse = true;
            ChainDrawer = DrawMTexChain;
        }
        else
        {
            MTexUse = false;
            ChainDrawer = DrawNoMTexChain;
        }
    }

    if (MTexUse)
    {
        // set up the multitexturing
        SelectTextureMTex (GL_MTexture1);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);

        SelectTextureMTex (GL_MTexture0);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

        SelectTextureMTex (GL_MTexture1);
        glEnable(GL_TEXTURE_2D);
    }
}

void R_ShutdownDraw (void)
{
    if (MTexUse)
    {
        // back to normal texturing
        glDisable(GL_TEXTURE_2D);

        SelectTextureMTex (GL_MTexture1);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

        SelectTextureMTex (GL_MTexture0);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    }
    MTexUse = false;
}


