/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// disable data conversion warnings

#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA

#ifdef _WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

void GL_SetPerspective (float fovx, float fovy);
void GL_BeginRendering (int *x, int *y, int *width, int *height);
void GL_EndRendering (void);


#ifdef _WIN32
// Function prototypes for the Texture Object Extension routines
typedef GLboolean (APIENTRY *ARETEXRESFUNCPTR)(GLsizei, const GLuint *, const GLboolean *);
typedef void (APIENTRY *BINDTEXFUNCPTR)(GLenum, GLuint);
typedef void (APIENTRY *DELTEXFUNCPTR)(GLsizei, const GLuint *);
typedef void (APIENTRY *GENTEXFUNCPTR)(GLsizei, GLuint *);
typedef GLboolean (APIENTRY *ISTEXFUNCPTR)(GLuint);
typedef void (APIENTRY *PRIORTEXFUNCPTR)(GLsizei, const GLuint *, const GLclampf *);
typedef void (APIENTRY *TEXSUBIMAGEPTR)(int, int, int, int, int, int, int, int, void *);

// faster texcoord functions
typedef void (APIENTRY *TEXCOORDFUNCPTR2F) (float, float);
typedef void (APIENTRY *TEXCOORDFUNCPTR2FV) (float*);

// faster vertex functions
typedef void (APIENTRY *VERTEXFUNCPTR2F) (float, float);
typedef void (APIENTRY *VERTEXFUNCPTR3F) (float, float, float);
typedef void (APIENTRY *VERTEXFUNCPTR3FV) (float*);

extern	BINDTEXFUNCPTR	bindTexFunc;
extern	DELTEXFUNCPTR	delTexFunc;
extern	TEXSUBIMAGEPTR	TexSubImage2DFunc;
#endif

extern	int		texture_extension_number;
extern	int		texture_mode;

int MHGL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int bitsperpixel);
int GL_LoadTexture (char *identifier, int *width, int *height, byte *data, qboolean mipmap, qboolean alpha);
int GL_FindTexture (char *identifier);

typedef struct
{
    float	x, y, z;
    float	s, t;
    float	r, g, b;
} glvert_t;

extern glvert_t glv;
extern float	gldepthmin, gldepthmax;
extern int		glx, gly, glwidth, glheight;

#ifdef _WIN32
extern	PROC glArrayElementEXT;
extern	PROC glColorPointerEXT;
extern	PROC glTexturePointerEXT;
extern	PROC glVertexPointerEXT;
#endif

// r_local.h -- private refresh defs
#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
// normalizing factor so player model works out to about
//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		    128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		    7
#define	SKYSIZE			    (1 << SKYSHIFT)
#define SKYMASK			    (SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01

void R_TimeRefresh_f (void);
void R_ReadPointFile_f (void);
gltexture_t *R_TextureAnimation (gltexture_t *base);

typedef struct surfcache_s
{
    struct surfcache_s	*next;
    struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
    int					lightadj[MAXLIGHTMAPS]; // checked for strobe flush
    int					dlight;
    int					size;		// including header
    unsigned			width;
    unsigned			height;		// DEBUG only needed for debug
    float				mipscale;
    struct gltexture_s	*texture;	// checked for animating textures
    byte				data[4];	// width*height elements
} surfcache_t;


typedef struct
{
    pixel_t		*surfdat;				// destination for generated surface
    int			rowbytes;				// destination logical width in bytes
    glmsurface_t	*surf;				// description for surface to generate
    fixed8_t	lightadj[MAXLIGHTMAPS];

    // adjust for lightmap levels for dynamic lighting
    gltexture_t	*texture;				// corrected for animating textures
    int			surfmip;				// mipmapped ratio of surface texels / world pixels
    int			surfwidth;				// in mipmapped texels
    int			surfheight;				// in mipmapped texels
} drawsurf_t;

extern float RTable[256];
extern float GTable[256];
extern float BTable[256];

extern float glowcos[17];
extern float glowsin[17];

extern float rotate_x;
extern float rotate_y;
extern float rotate_z;

//====================================================

extern  float       lerp_mody;

extern	int			global_worldtype;
extern	int			worldtype_found;

extern	qboolean	hlbsp;
extern	qboolean	qsg30;
extern	qboolean	inwater;

extern	int			gl_useskybox;
extern	int			gl_fogenabled;
extern  int			gl_useclight;

extern	entity_t	r_worldentity;
extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	vec3_t		lightspot;
extern	entity_t	*currententity;
extern	float		frametime;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	mplane_t	frustum[4];
extern  GLfloat     fog_colours[4];
extern	int			c_brush_polys, c_alias_polys;

//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	gltexture_t	*r_notexture_mip;
extern	int			d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	int			cnttextures[2];
extern	int			particletexture;
extern	int			playertextures;

extern	int			skytexturenum;		// index in cl.loadmodel, not gl texture object
extern	int			notexture_num;

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	_r_skybox;
extern  cvar_t  _r_persist;
extern	cvar_t	r_detailsky;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;
extern  cvar_t  r_autowatertrans;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_mirroralpha;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;

extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_poly;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_keeptjunctions;
extern	cvar_t	gl_reporttjunctions;
extern	cvar_t	gl_fastbutugly;
extern  cvar_t  gl_multitexture;
extern	cvar_t	gl_nocolors;
extern	cvar_t	gl_doubleeyes;
extern  cvar_t  gl_dlradmult;
extern	cvar_t	gl_aliasglow;
extern	cvar_t	gl_aliastrans;
extern	cvar_t	gl_aliaslightlerp;
extern  cvar_t  gl_fullbright;
extern  cvar_t  gl_fullbright_alias;
extern	cvar_t	gl_colouredlight;
extern  cvar_t	gl_vertexlight;
extern	cvar_t	gl_subdivide_size;
extern  cvar_t  gl_texscale;

extern	int		gl_lightmap_format;
extern	int		gl_solid_format;
extern	int		gl_alpha_format;

extern	cvar_t	gl_max_size;
extern	cvar_t	gl_playermip;

extern	float	r_world_matrix[16];
extern	float   r_base_world_matrix[16];
extern  float   r_projection_matrix[16];

extern  cvar_t  gl_skyfog;
extern  cvar_t  gl_fogstart;
extern  cvar_t  gl_fogend;

extern	cvar_t	sky_rotate_x;
extern	cvar_t	sky_rotate_y;
extern	cvar_t	sky_rotate_z;

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

extern byte *image_rgba;
extern FILE *file32bit;

void EmitWaterPolys (glmsurface_t *fa, float temp_wave, float trans_lev, float warp_factor);
void EmitSkyPolys (glmsurface_t *fa);
void EmitSimpleSkyPolys (glmsurface_t *fa);

int LoadTGA (int *width, int *height, qboolean complain, qboolean blah);
int LoadJPG (int *width, int *height, qboolean complain, qboolean blah);
int LoadPCX (int *width, int *height, qboolean complain, qboolean use_fake_alpha);

void GL_BuildVertexLightColour (glmsurface_t *fa);
void GL_BuildVertexLightWhite (glmsurface_t *fa);

// this never did have a prototype, even the the original quake source - weird...
void GL_SubdivideSurface (glmsurface_t *fa, int subdiv_size);
void R_DrawFullBrightPolys (glmsurface_t *s);
void R_TranslatePlayerSkin (int playernum);
void GL_Bind (int texnum);

void R_DrawSkyBox (void);
void R_ClearSkyBox (void);
void R_DrawOldSkyBox (void);
void R_DrawOldSkyboxChain (glmsurface_t *s);

