/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// alias model loading functions - seperated out to make gl_model.c more readable

#include "quakedef.h"

aliashdr_t	*pheader;
stvert_t	stverts[MAXALIASVERTS];
mtriangle_t	triangles[MAXALIASTRIS];

// a pose is a single set of vertexes.  a frame may be
// an animating sequence of poses
trivertx_t	*poseverts[MAXALIASFRAMES];
trivert2	posenorms[MAXALIASFRAMES][MAXALIASVERTS]; // muff - temp holding space for the normals
int			posenum;

// proper bounding box for alias models - courtesy lordhavoc
vec3_t aliasbboxmin;
vec3_t aliasbboxmax;

typedef struct
{
    int v[3];
    vec3_t n;
} temptris_t;

temptris_t  *temptris;
qboolean    load_fbc;

//originally LordHavoc code - hacked about by muff to completely save re-writing it from scratch
// mh - possibly for q2 only???
void Mod_ConvertAliasVerts (int numverts, int numtris, vec3_t scale, vec3_t translate, trivertx_t *v, trivert2 *out)
{
    int     i, j;
    vec3_t  t1, t2;

    struct
    {
        vec3_t  v;
        vec3_t  n;
        int     count;
    } tempvert[MAXALIASVERTS];

    temptris_t *tris;

    // decompress vertices
    for (i = 0; i < numverts; i++)
    {
        tempvert[i].v[0] = v[i].v[0] * scale[0] + translate[0];
        tempvert[i].v[1] = v[i].v[1] * scale[1] + translate[1];
        tempvert[i].v[2] = v[i].v[2] * scale[2] + translate[2];
        tempvert[i].n[0] = tempvert[i].n[1] = tempvert[i].n[2] = 0;
        tempvert[i].count = 0;

        // update bounding box
        if (tempvert[i].v[0] < aliasbboxmin[0]) aliasbboxmin[0] = tempvert[i].v[0];
        if (tempvert[i].v[1] < aliasbboxmin[1]) aliasbboxmin[1] = tempvert[i].v[1];
        if (tempvert[i].v[2] < aliasbboxmin[2]) aliasbboxmin[2] = tempvert[i].v[2];
        if (tempvert[i].v[0] > aliasbboxmax[0]) aliasbboxmax[0] = tempvert[i].v[0];
        if (tempvert[i].v[1] > aliasbboxmax[1]) aliasbboxmax[1] = tempvert[i].v[1];
        if (tempvert[i].v[2] > aliasbboxmax[2]) aliasbboxmax[2] = tempvert[i].v[2];
    }

    // calculate surface normals
    tris = temptris;

    for (i = 0; i < numtris; i++)
    {
        //do the calculations
        VectorSubtract(tempvert[tris->v[0]].v, tempvert[tris->v[1]].v, t1);
        VectorSubtract(tempvert[tris->v[2]].v, tempvert[tris->v[1]].v, t2);
        CrossProduct(t1, t2, tris->n);
        VectorNormalize(tris->n);

        // add surface normal to vertices
        for (j = 0; j < 3; j++)
        {
            VectorAdd(tris->n, tempvert[tris->v[j]].n, tempvert[tris->v[j]].n);
            tempvert[tris->v[j]].count++;
        }
        tris++;
    }

    // average normals and write out 1.7bit format
    for (i = 0; i < /*pheader->*/numverts; i++)				// muff -  fixed
    {
        VectorNormalize(tempvert[i].n);

        out[i].n[0] = (signed char) (tempvert[i].n[0] * 127.0);
        out[i].n[1] = (signed char) (tempvert[i].n[1] * 127.0);
        out[i].n[2] = (signed char) (tempvert[i].n[2] * 127.0);
    }
}

/*
=================
Mod_LoadAliasFrame
=================
*/
void * Mod_LoadAliasFrame (void * pin, maliasframedesc_t *frame)
{
    trivertx_t		*pinframe;
    int				i, j;
    daliasframe_t	*pdaliasframe;
    trivert2		*nin;

    pdaliasframe = (daliasframe_t *)pin;

    strcpy (frame->name, pdaliasframe->name);
    frame->firstpose = posenum;
    frame->numposes = 1;

    for (i=0 ; i<3 ; i++)
    {
        // these are byte values, so we don't have to worry about
        // endianness
        frame->bboxmin.v[i] = pdaliasframe->bboxmin.v[i];
        frame->bboxmax.v[i] = pdaliasframe->bboxmax.v[i];	// bug - was min

        // LordHavoc: doing proper bbox for model
        if (frame->bboxmin.v[i] < aliasbboxmin[i]) aliasbboxmin[i] = frame->bboxmin.v[i];
        if (frame->bboxmax.v[i] > aliasbboxmax[i]) aliasbboxmax[i] = frame->bboxmax.v[i];
    }

    pinframe = (trivertx_t *)(pdaliasframe + 1);

    poseverts[posenum] = pinframe;

    Mod_ConvertAliasVerts(pheader->numverts, pheader->numtris, pheader->scale, pheader->scale_origin, pinframe, (void *)((int) pheader + pheader->normdata + sizeof(trivert2) * pheader->numverts * posenum) );

    // there is bound to be an easier way, but this makes sense
    // basically copying all the normal data to temparea, so that is can be shuffled in gl_mesh.c
    nin = (void *)((int) pheader + pheader->normdata + sizeof(trivert2) * pheader->numverts * posenum);

    for (j=0; j <pheader->numverts ; j++)
    {
        posenorms[posenum][j].n[0] = nin->n[0];
        posenorms[posenum][j].n[1] = nin->n[1];
        posenorms[posenum][j].n[2] = nin->n[2];
        nin ++ ;
    }
    posenum++;

    pinframe += pheader->numverts;

    return (void *)pinframe;
}

/*
=================
Mod_LoadAliasGroup
=================
*/
void *Mod_LoadAliasGroup (void * pin,  maliasframedesc_t *frame)
{
    daliasgroup_t		*pingroup;
    int					i, j, numframes;
    daliasinterval_t	*pin_intervals;
    void				*ptemp;
    trivert2		    *nin;

    pingroup = (daliasgroup_t *)pin;

    numframes = LittleLong (pingroup->numframes);

    frame->firstpose = posenum;
    frame->numposes = numframes;

    for (i=0 ; i<3 ; i++)
    {
        // these are byte values, so we don't have to worry about endianness
        frame->bboxmin.v[i] = pingroup->bboxmin.v[i];
        frame->bboxmax.v[i] = pingroup->bboxmax.v[i];	// bug - was min

        // LordHavoc: doing proper bbox for model
        if (frame->bboxmin.v[i] < aliasbboxmin[i]) aliasbboxmin[i] = frame->bboxmin.v[i];
        if (frame->bboxmax.v[i] > aliasbboxmax[i]) aliasbboxmax[i] = frame->bboxmax.v[i];
    }
    pin_intervals = (daliasinterval_t *)(pingroup + 1);

    frame->interval = LittleFloat (pin_intervals->interval);

    pin_intervals += numframes;

    ptemp = (void *)pin_intervals;

    for (i=0 ; i<numframes ; i++)
    {
        poseverts[posenum] = (trivertx_t *)((daliasframe_t *)ptemp + 1);

        Mod_ConvertAliasVerts(pheader->numverts, pheader->numtris, pheader->scale, pheader->scale_origin, (void *)((daliasframe_t *)ptemp + 1), (void *)((int) pheader + pheader->normdata + sizeof(trivert2) * pheader->numverts * posenum) );//, (void *)((int) pheader + pheader->vertindices2));

        // same ugly hack as the single frame one above
        nin = (void *)((int) pheader + pheader->normdata + sizeof(trivert2) * pheader->numverts * posenum);

        for (j=0; j <pheader->numverts ; j++)
        {
            posenorms[posenum][j].n[0] = nin->n[0];
            posenorms[posenum][j].n[1] = nin->n[1];
            posenorms[posenum][j].n[2] = nin->n[2];
            nin ++ ;
        }
        posenum++;

        ptemp = (trivertx_t *)((daliasframe_t *)ptemp + 1) + pheader->numverts;
    }
    return ptemp;
}

//=========================================================

/*
=================
Mod_FloodFillSkin

Fill background pixels so mipmapping doesn't have haloes - Ed
=================
*/

typedef struct
{
    short		x, y;
} floodfill_t;

extern unsigned d_8to24table[];

// must be a power of 2
#define FLOODFILL_FIFO_SIZE 0x1000
#define FLOODFILL_FIFO_MASK (FLOODFILL_FIFO_SIZE - 1)

#define FLOODFILL_STEP( off, dx, dy ) \
{ \
	if (pos[off] == fillcolor) \
	{ \
		pos[off] = 255; \
		fifo[inpt].x = x + (dx), fifo[inpt].y = y + (dy); \
		inpt = (inpt + 1) & FLOODFILL_FIFO_MASK; \
	} \
	else if (pos[off] != 255) fdc = pos[off]; \
}

void Mod_FloodFillSkin( byte *skin, int skinwidth, int skinheight )
{
    byte				fillcolor = *skin; // assume this is the pixel to fill
    floodfill_t			fifo[FLOODFILL_FIFO_SIZE];
    int					inpt = 0, outpt = 0;
    int					filledcolor = -1;
    int					i;

    if (filledcolor == -1)
    {
        filledcolor = 0;

        // attempt to find opaque black
        for (i = 0; i < 256; ++i)
        {
            if (d_8to24table[i] == (255 << 0)) // alpha 1.0
            {
                filledcolor = i;
                break;
            }
        }
    }

    // can't fill to filled color or to transparent color (used as visited marker)
    if ((fillcolor == filledcolor) || (fillcolor == 255)) return;

    fifo[inpt].x = 0, fifo[inpt].y = 0;
    inpt = (inpt + 1) & FLOODFILL_FIFO_MASK;

    while (outpt != inpt)
    {
        int			x = fifo[outpt].x, y = fifo[outpt].y;
        int			fdc = filledcolor;
        byte		*pos = &skin[x + skinwidth * y];

        outpt = (outpt + 1) & FLOODFILL_FIFO_MASK;

        if (x > 0)				FLOODFILL_STEP( -1, -1, 0 );
        if (x < skinwidth - 1)	FLOODFILL_STEP( 1, 1, 0 );
        if (y > 0)				FLOODFILL_STEP( -skinwidth, 0, -1 );
        if (y < skinheight - 1)	FLOODFILL_STEP( skinwidth, 0, 1 );
        skin[x + skinwidth * y] = fdc;
    }
}

/*
===============
Mod_LoadAllSkins
===============
*/
byte *player_texels;

void *Mod_LoadAllSkins (int numskins, daliasskintype_t *pskintype)
{
    int		i, j, k;
    char	name[32];
    int		s;
    byte	*skin;
    byte	*texels;
    daliasskingroup_t		*pinskingroup;
    int		groupskins;
    daliasskininterval_t	*pinskinintervals;
    char    tex_name[1024];
    int     pixels;
    int     oldwidth, oldheight;

    skin = (byte *)(pskintype + 1);

    if (numskins < 1 || numskins > MAX_SKINS)
        Sys_Error ("Mod_LoadAliasModel: Invalid # of skins: %d\n", numskins);

    s = pheader->skinwidth * pheader->skinheight;

    for (i=0 ; i<numskins ; i++)
    {
        if (pskintype->type == ALIAS_SKIN_SINGLE)
        {
            Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

            // save 8 bit texels for the player model to remap - we shouldn't have to do this
            // anymore. but the engine crashes if we don't - one to track down...
            player_texels = Hunk_AllocName(s, loadname);

            pheader->texels[i] = player_texels - (byte *)pheader;

            // save the player model texels again seperately cos the above is bugged
            // this is more of a nasty hack of a workaround than a fix, but if it works...
            if (pheader->is_playermodel)
            {
                memcpy (player_texels, (byte *)(pskintype + 1), s);
            }
            sprintf (name, "%s_%i", loadmodel->name, i);

            // save skin width and height cos we don't want gl_loadtexture to mess them up
            oldwidth = pheader->skinwidth;
            oldheight = pheader->skinheight;

            pheader->gl_texturenum[i][0] =
                pheader->gl_texturenum[i][1] =
                    pheader->gl_texturenum[i][2] =
                        pheader->gl_texturenum[i][3] = GL_LoadTexture (name, &pheader->skinwidth,
                                &pheader->skinheight, (byte *)(pskintype + 1), true, false);

            // restore skin width and height in case they've been messed up by gl_loadtexture
            pheader->skinwidth = oldwidth;
            pheader->skinheight = oldheight;

            // skins aren't stored mipmapped, so don't do the old / 64 * 85
            pixels = pheader->skinwidth * pheader->skinheight;

            // check for fullbrights in model skins - this is ugly...
            if (Mod_FindFullBrights ((byte *)(pskintype + 1), pixels) && load_fbc)
            {
                // convert any non fullbright pixel to fully transparent
                Mod_ConvertFullbrights ((byte *)(pskintype + 1), pixels);

                // get a new name for the fullbright mask to avoid cache mismatches
                sprintf (tex_name, "fbrm_%s", name);

                // save skin width and height cos we don't want gl_loadtexture to mess them up
                oldwidth = pheader->skinwidth;
                oldheight = pheader->skinheight;

                // load the fullbright pixels version of the texture
                pheader->fullbright[i][0] =
                    pheader->fullbright[i][1] =
                        pheader->fullbright[i][2] =
                            pheader->fullbright[i][3] = GL_LoadTexture (tex_name, &pheader->skinwidth,
                                                        &pheader->skinheight, (byte *)(pskintype + 1), true, true);

                // restore skin width and height in case they've been messed up by gl_loadtexture
                pheader->skinwidth = oldwidth;
                pheader->skinheight = oldheight;
            }
            else pheader->fullbright[i][0] =
                    pheader->fullbright[i][1] =
                        pheader->fullbright[i][2] =
                            pheader->fullbright[i][3] = -1;	// because 0 is a potentially valid texture number

            pskintype = (daliasskintype_t *)((byte *)(pskintype+1) + s);
        }
        else
        {
            // animating skin group.  yuck.
            pskintype++;
            pinskingroup = (daliasskingroup_t *)pskintype;
            groupskins = LittleLong (pinskingroup->numskins);
            pinskinintervals = (daliasskininterval_t *)(pinskingroup + 1);

            pskintype = (void *)(pinskinintervals + groupskins);

            for (j=0 ; j<groupskins ; j++)
            {
                Mod_FloodFillSkin (skin, pheader->skinwidth, pheader->skinheight);

                if (j == 0)
                {
                    // save 8 bit texels for the player model to remap - we shouldn't have to do
                    // this anymore. but the engine crashes if we don't - one to track down...
                    player_texels = Hunk_AllocName(s, loadname);

                    pheader->texels[i] = player_texels - (byte *)pheader;

                    // save the player model texels again seperately cos the above is bugged
                    // this is more of a nasty hack of a workaround than a fix, but if it works...
                    if (pheader->is_playermodel)
                    {
                        memcpy (player_texels, (byte *)(pskintype), s);
                    }
                }
                sprintf (name, "%s_%i_%i", loadmodel->name, i,j);

                // save skin width and height cos we don't want gl_loadtexture to mess them up
                oldwidth = pheader->skinwidth;
                oldheight = pheader->skinheight;

                pheader->gl_texturenum[i][j&3] =
                    GL_LoadTexture (name, &pheader->skinwidth,
                                    &pheader->skinheight, (byte *)(pskintype), true, false);

                // restore skin width and height in case they've been messed up by gl_loadtexture
                pheader->skinwidth = oldwidth;
                pheader->skinheight = oldheight;

                pixels = pheader->skinheight * pheader->skinwidth;

                if (Mod_FindFullBrights ((byte *)(pskintype), pixels) && load_fbc)
                {
                    // convert any non fullbright pixel to fully transparent
                    Mod_ConvertFullbrights ((byte *)(pskintype), pixels);

                    // get a new name for the fullbright mask to avoid cache mismatches
                    sprintf (tex_name, "fbrm_%s", name);

                    // save skin width and height cos we don't want gl_loadtexture to mess them up
                    oldwidth = pheader->skinwidth;
                    oldheight = pheader->skinheight;

                    // load the fullbright pixels version of the texture
                    pheader->fullbright[i][j&3] = GL_LoadTexture (tex_name,
                                                  &pheader->skinwidth, &pheader->skinheight,
                                                  (byte *)(pskintype + 1), true, true);

                    // restore skin width and height in case they've been messed up by
                    // gl_loadtexture
                    pheader->skinwidth = oldwidth;
                    pheader->skinheight = oldheight;
                }
                else pheader->fullbright[i][j&3] = -1;

                pskintype = (daliasskintype_t *)((byte *)(pskintype) + s);
            }

            k = j;

            for (/* */; j < 4; j++)
            {
                pheader->gl_texturenum[i][j & 3] = pheader->gl_texturenum[i][j - k];
                pheader->fullbright[i][j & 3] = pheader->fullbright[i][j - k];
            }
        }
    }

    return (void *)pskintype;
}


//=========================================================================

/*
=================
Mod_LoadAliasModel
=================
*/
void Mod_LoadAliasModel (model_t *mod, void *buffer)
{
    int					i, j;
    mdl_t				*pinmodel;
    stvert_t			*pinstverts;
    dtriangle_t			*pintriangles;
    int					version, numframes;
    int					size;
    daliasframetype_t	*pframetype;
    daliasskintype_t	*pskintype;
    int					start, end, total;

    // LordHavoc:
    unsigned short		*poutvertindices;
    float				*pouttexcoords;
    temptris_t			*tris;

    start = Hunk_LowMark ();

    // vicious!  memory leak!!! it's better to just malloc and free for every model
    // if (!temptris) temptris = malloc(sizeof(temptris_t) * MAXALIASTRIS);
    temptris = malloc(sizeof(temptris_t) * MAXALIASTRIS);

    pinmodel = (mdl_t *)buffer;

    version = LittleLong (pinmodel->version);

    if (version != ALIAS_VERSION)
        Sys_Error ("%s has wrong version number (%i should be %i)",
                   mod->name, version, ALIAS_VERSION);

    // allocate space for a working header, plus all the data except the frames,
    // skin and group info
    // size = sizeof (aliashdr_t) + (LittleLong (pinmodel->numframes) - 1) * sizeof (pheader->frames[0]);
    size = sizeof (aliashdr_t);
    size += LittleLong (pinmodel->numframes) * (sizeof(trivertx_t) * LittleLong (pinmodel->numverts));
    size += LittleLong (pinmodel->numtris) * sizeof(unsigned short[3]);
    size += LittleLong (pinmodel->numframes) * (sizeof(trivert2) * LittleLong (pinmodel->numverts));

    if (size < 256 || size > 4194304) // weird size
        Host_Error ("model %s is corrupt\n", mod->name);

    pheader = Hunk_AllocName (size, loadname);

    mod->flags = LittleLong (pinmodel->flags);

    // endian-adjust and copy the data, starting with the alias model header
    pheader->boundingradius = LittleFloat (pinmodel->boundingradius);
    pheader->numskins = LittleLong (pinmodel->numskins);
    pheader->skinwidth = LittleLong (pinmodel->skinwidth);
    pheader->skinheight = LittleLong (pinmodel->skinheight);

    if (pheader->skinheight > MAX_LBM_HEIGHT)
        Sys_Error ("model %s has a skin taller than %d", mod->name,
                   MAX_LBM_HEIGHT);

    pheader->numverts = LittleLong (pinmodel->numverts);

    if (pheader->numverts <= 0)
        Sys_Error ("model %s has no vertices", mod->name);

    if (pheader->numverts > MAXALIASVERTS)
        Sys_Error ("model %s has too many vertices", mod->name);

    pheader->numtris = LittleLong (pinmodel->numtris);

    if (pheader->numtris <= 0)
        Sys_Error ("model %s has no triangles", mod->name);

    pheader->numframes = LittleLong (pinmodel->numframes);
    numframes = pheader->numframes;

    if (numframes < 1)
        Sys_Error ("Mod_LoadAliasModel: Invalid # of frames: %d\n", numframes);

    pheader->size = LittleFloat (pinmodel->size) * ALIAS_BASE_SIZE_RATIO;
    mod->synctype = LittleLong (pinmodel->synctype);
    mod->numframes = pheader->numframes;

    for (i=0 ; i<3 ; i++)
    {
        pheader->scale[i] = LittleFloat (pinmodel->scale[i]);
        pheader->scale_origin[i] = LittleFloat (pinmodel->scale_origin[i]);
        pheader->eyeposition[i] = LittleFloat (pinmodel->eyeposition[i]);
    }

    // MFAH - check for fullbright drawing (some models should be fullbright)
    // this allows for 3 levels of fixed brightness for most of the alias models,
    // resulting in a somewhat significant speeding up of alias model drawing.
    // (ie. no need for an extensive r_lightpoint check)
    // have to check these before loading the skins cos mdls which are fullbright
    // already should not use fullbright colour masks!
    if ((!strncmp (mod->name, "progs/flame",11))      ||
            (!strcmp  (mod->name, "progs/laser.mdl"))     ||
            (!strcmp  (mod->name, "progs/w_spike.mdl"))   ||
            (!strcmp  (mod->name, "progs/v_spike.mdl"))   ||
            (!strcmp  (mod->name, "progs/k_spike.mdl"))   ||
            (!strcmp  (mod->name, "progs/candle.mdl"))	  ||
            (!strcmp  (mod->name, "progs/fireball.mdl"))  ||
            (!strcmp  (mod->name, "progs/lantern.mdl"))   ||
            (!strncmp (mod->name, "progs/end",9))         ||
            (!strcmp  (mod->name, "progs/lspike.mdl"))    ||
            (!strcmp  (mod->name, "progs/lasrspik.mdl"))  ||
            (!strcmp  (mod->name, "progs/invulner.mdl"))  ||
            (!strcmp  (mod->name, "progs/lavaball.mdl"))  ||
            (!strcmp  (mod->name, "progs/plasma.mdl"))    ||
            (!strcmp  (mod->name, "progs/s_light.mdl")))
        pheader->is_fullbright = 160;
    else if ((!strncmp (mod->name, "progs/w_g_",10))  ||
             (!strncmp (mod->name, "progs/w_s_",10))       ||
             (!strncmp (mod->name, "progs/b_g_",10))       ||
             (!strncmp (mod->name, "progs/b_s_",10))       ||
             (!strncmp (mod->name, "progs/m_g_",10))       ||
             (!strncmp (mod->name, "progs/m_s_",10))       ||
             (!strcmp  (mod->name, "progs/missile.mdl"))   ||
             (!strcmp  (mod->name, "progs/quaddama.mdl"))  ||
             (!strcmp  (mod->name, "progs/invisibl.mdl"))  ||
             (!strcmp  (mod->name, "progs/lavarock.mdl"))  ||
             (!strcmp  (mod->name, "progs/empathy.mdl"))   ||
             (!strncmp (mod->name, "progs/g_",8))          ||
             (!strncmp (mod->name, "progs/bolt",10)))
        pheader->is_fullbright = 128;
    else if ((!strncmp (mod->name, "progs/gib",9))    ||
             (!strcmp  (mod->name, "progs/grenade.mdl"))   ||
             (!strcmp  (mod->name, "progs/zom_gib.mdl"))   ||
             (!strcmp  (mod->name, "progs/armor.mdl"))     ||
             (!strcmp  (mod->name, "progs/beam.mdl"))	  ||
             (!strcmp  (mod->name, "progs/horn.mdl"))	  ||
             (!strcmp  (mod->name, "progs/suit.mdl"))	  ||
             (!strcmp  (mod->name, "progs/wetsuit.mdl"))   ||
             (!strcmp  (mod->name, "progs/proxbomb.mdl"))  ||
             (!strcmp  (mod->name, "progs/backpack.mdl")))
        pheader->is_fullbright = 96;
    else pheader->is_fullbright = 0;

    if (pheader->is_fullbright)
         load_fbc = false;
    else load_fbc = true;

    // identify the player model so we can interpolate it
    if (!strcmp  (mod->name, "progs/player.mdl"))
         pheader->is_playermodel = true;
    else pheader->is_playermodel = false;

    // load the skins
    pskintype = (daliasskintype_t *)&pinmodel[1];
    pskintype = (daliasskintype_t *)Mod_LoadAllSkins (pheader->numskins, pskintype);

    // load base s and t vertices
    pinstverts = (stvert_t *)pskintype;
    pouttexcoords = (float *)&pheader->frames[numframes];
    pheader->texcoords = (int) pouttexcoords - (int) pheader;

    for (i=0 ; i<pheader->numverts ; i++)
    {
        stverts[i].onseam = LittleLong (pinstverts[i].onseam);
        stverts[i].s = LittleLong (pinstverts[i].s);
        stverts[i].t = LittleLong (pinstverts[i].t);
    }

    // load triangle lists
    pintriangles = (dtriangle_t *)&pinstverts[pheader->numverts];
    poutvertindices = (unsigned short *)&pouttexcoords[pheader->numverts*4];
    pheader->vertindices = (int) poutvertindices - (int) pheader;

    // we double load the triangle list here, to put the triangles in a seperate memory area for
    // the moment...hey, it works - muff
    tris = temptris;

    for (i=0 ; i<pheader->numtris ; i++)
    {
        triangles[i].facesfront = LittleLong (pintriangles[i].facesfront);

        for (j=0 ; j<3 ; j++)
        {
            triangles[i].vertindex[j] =	LittleLong (pintriangles[i].vertindex[j]);

            //load the triangles into tris[] as well so we can calculate normals
            tris->v[j] = LittleLong (pintriangles[i].vertindex[j]);
        }
        tris++;
    }

    // load the frames
    posenum = 0;
    pheader->posedata = (int) poutvertindices - (int) pheader; // LH

    // muff - set pointer for the vertnorms we'll calculate in a moment
    pheader->normdata = pheader->posedata + LittleLong (pinmodel->numframes) * (sizeof(trivertx_t) * LittleLong (pinmodel->numverts));

    pframetype = (daliasframetype_t *)&pintriangles[pheader->numtris];

    aliasbboxmin[0] = aliasbboxmin[1] = aliasbboxmin[2] = 1000;
    aliasbboxmax[0] = aliasbboxmax[1] = aliasbboxmax[2] = -1000;

    for (i=0 ; i<numframes ; i++)
    {
        aliasframetype_t	frametype;

        frametype = LittleLong (pframetype->type);

        if (frametype == ALIAS_SINGLE)
            pframetype = (daliasframetype_t *) Mod_LoadAliasFrame (pframetype + 1, &pheader->frames[i]);
        else
            pframetype = (daliasframetype_t *) Mod_LoadAliasGroup (pframetype + 1, &pheader->frames[i]);
    }

    pheader->numposes = posenum;

    mod->type = mod_alias;

    // LordHavoc: fixed model bbox - was //FIXME: do this right
    for (j = 0; j < 3; j++)
    {
        mod->mins[j] = aliasbboxmin[j] * pheader->scale[j] + pheader->scale_origin[j];
        mod->maxs[j] = aliasbboxmax[j] * pheader->scale[j] + pheader->scale_origin[j];
    }

    // build the draw lists - the engine crashes during this process if we don't save the skin
    // texels
    GL_MakeAliasModelDisplayLists (mod, pheader);

    // special glow behaviour for torches, missiles, bolts & powerups
    pheader->is_missile = false;
    pheader->is_torch = false;
    pheader->is_powerup = false;
    pheader->is_bolt = false;

    // MFAH - check for model transparency and glowmapping here
    // 27-6-00 added rogue and hipnotic models
    // the value of pheader->is_glowmap is either 0 (no glow) or an index into the quake palette
    // never really got round to doing that palette index bit, did i?
    if (!strncmp (mod->name, "progs/flame",11))
    {
        pheader->is_glowmap = 236;
        pheader->is_torch = true;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strncmp (mod->name, "progs/bolt",10))
    {
        pheader->is_bolt = true;
        pheader->is_glowmap = 244;
        pheader->glow_rgb[0] = 0.05;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.2;
    }
    else if (!strcmp  (mod->name, "progs/missile.mdl"))
    {
        pheader->is_glowmap = 236;
        pheader->is_missile = true;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strcmp  (mod->name, "progs/laser.mdl"))
    {
        pheader->is_glowmap = 251;
        pheader->glow_rgb[0] = 0.2;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.025;
    }
    else if (!strcmp  (mod->name, "progs/w_spike.mdl"))
    {
        pheader->is_glowmap = 63;
        pheader->glow_rgb[0] = 0.025;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.025;
    }
    else if (!strcmp  (mod->name, "progs/v_spike.mdl"))
    {
        pheader->is_glowmap = 147;
        pheader->glow_rgb[0] = 0.2;
        pheader->glow_rgb[1] = 0.025;
        pheader->glow_rgb[2] = 0.2;
    }
    else if (!strcmp  (mod->name, "progs/k_spike.mdl"))
    {
        pheader->is_glowmap = 236;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strcmp  (mod->name, "progs/beam.mdl"))
    {
        pheader->is_glowmap = 236;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strcmp  (mod->name, "progs/candle.mdl"))
    {
        pheader->is_glowmap = 236;
        pheader->is_torch = true;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strcmp  (mod->name, "progs/lantern.mdl"))
    {
        pheader->is_glowmap = 236;
        pheader->is_torch = true;
        pheader->glow_rgb[0] = 0.4;
        pheader->glow_rgb[1] = 0.2;
        pheader->glow_rgb[2] = 0.1;
    }
    else if (!strcmp  (mod->name, "progs/lspike.mdl"))
    {
        pheader->is_glowmap = 251;
        pheader->glow_rgb[0] = 0.2;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.025;
    }
    else if (!strcmp  (mod->name, "progs/lasrspik.mdl"))
    {
        pheader->is_glowmap = 251;
        pheader->glow_rgb[0] = 0.2;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.025;
    }
    else if (!strcmp  (mod->name, "progs/plasma.mdl"))
    {
        pheader->is_glowmap = 244;
        pheader->glow_rgb[0] = 0.05;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.2;
    }
    else if (!strcmp  (mod->name, "progs/s_light.mdl"))
    {
        // shambler lightning attack - we need to either
        // (1) disable the glow (cop-out), or
        // (2) find a way to move the origin up
        pheader->is_glowmap = 244;
        pheader->glow_rgb[0] = 0.05;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.2;
    }
    else if (!strcmp  (mod->name, "progs/quaddama.mdl"))
    {
        pheader->is_powerup = true;
        pheader->is_glowmap = 244;
        pheader->glow_rgb[0] = 0.05;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.2;
    }
    else if (!strcmp  (mod->name, "progs/invulner.mdl"))
    {
        pheader->is_powerup = true;
        pheader->is_glowmap = 251;
        pheader->glow_rgb[0] = 0.2;
        pheader->glow_rgb[1] = 0.05;
        pheader->glow_rgb[2] = 0.025;
    }
    else
    {
        pheader->is_glowmap = 0;
        pheader->glow_rgb[0] = 0;
        pheader->glow_rgb[1] = 0;
        pheader->glow_rgb[2] = 0;
    }

    // MFAH - check for model transparency
    if (!strncmp (mod->name, "progs/bolt",10) || !strcmp  (mod->name, "progs/s_light.mdl"))
    {
        // bolts are fullbright & therefore harder to see the transparency
        pheader->trans_level = 0.25;
    }
    else if (!strcmp (mod->name, "progs/wizard.mdl") || !strcmp (mod->name, "progs/h_wizard.mdl"))
    {
        // wizards are harder to see when transparent, so make them more solid
        pheader->trans_level = 0.75;
    }
    else if ((!strcmp (mod->name, "progs/laser.mdl")) ||
             (!strcmp (mod->name, "progs/w_spike.mdl")) ||
             (!strcmp (mod->name, "progs/v_spike.mdl")) ||
             (!strcmp (mod->name, "progs/k_spike.mdl")))
        pheader->trans_level = 0.5;
    else pheader->trans_level = 1.0;

    // mfah - check for shadow casting by model name
    if (pheader->trans_level < 1.0)
        pheader->cast_shadow = 0;	// no shadow if transparent
    else if (!strncmp (mod->name, "progs/v_",8))
        pheader->cast_shadow = 0;	// don't use a shadow on viewmodels cos they're not suitable
    else pheader->cast_shadow = 1;

    // mfah - check for interpolation - i've changed my mind about the viewmodels.
    // there's sometimes a weird effect with the sng when switching from active to inactive
    // in mid-spin - this is a QC bug - it should go through the full spin cycle before becoming
    // inactive rather than doing so straight away.  unfortunately, this is more easily done in
    // qc than the engine, so we'll just jump straight to frame 0 without interpolating instead.
    // interpolate 0 is used to identify the sng in gl_aliasmodel.c
    if (!strcmp (mod->name, "progs/v_nail2.mdl"))
        pheader->interpolate = 0;	// identify the sng
    else pheader->interpolate = 1;

    // don't do bounding box culling on view models - this will also affect the shalrath
    // missile, but that doesn't really matter i suppose - cheap and nasty hack
    if (!strncmp (mod->name, "progs/v_", 8))
        pheader->nocull = 1;
    else pheader->nocull = 0;

    // move the complete, relocatable alias model to the cache
    end = Hunk_LowMark ();
    total = end - start;

    Cache_Alloc (&mod->cache, total, loadname);
    if (!mod->cache.data) return;
    memcpy (mod->cache.data, pheader, total);
    Hunk_FreeToLowMark (start);
    free (temptris);
}
