/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// miscellaneous edict manipulation routines - most of these are not actually used in the
// engine just yet, but they're always handy to have for future expansion.

// this stuff is in an extremely unfinished state right now.  i had code to do most of it
// in an earlier engine, but have to spend some time looking for it.  yuck.

#include "quakedef.h"


/*
PR_FuncNum

Returns the internal number of a QuakeC function - handy to have if we want to
change stuff like think functions on the fly within the engine
*/
dfunction_t *ED_FindFunction (char *name);

int PR_FuncNum (char *fname)
{
    int i;
    dfunction_t	*findfunc;

    findfunc = ED_FindFunction (fname);

    for (i = 0; i < progs->numfunctions; i++)
        if (pr_functions[i].first_statement == findfunc->first_statement) break;

    if (i == progs->numfunctions)
        return 0;	// not found

    return i;
}


/*
PR_FindClassname

Get the classname of any given edict - there is no really convenient direct way to do
this, so we have to go about it the long way.  Fixme: modify this to find any given
field
*/
void PR_FindClassname (edict_t *ed, char *classname)
{
    // should work - we'll have to test it
    classname = pr_strings + ed->v.classname;
}


/*
PR_PersistGibs

just call it from any function that is called once per frame and you'll
have persistent gibs.
*/
int Num_SubRemove, Num_SubNull;

void PR_NumGibSubs (void)
{
    Num_SubRemove = PR_FuncNum ("SUB_Remove");
    Num_SubNull = PR_FuncNum ("SUB_Null");
}


char *PR_UglyValueString (etype_t type, eval_t *val);

// you can use a similar function to make backpacks stay around for ever, ie. check if the
// think is sub_remove and the touch is backpacktouch... and for various other stuff.
void PR_PersistGibs (void)
{
    edict_t *ent;
    int j;

    for (j = 0; j < sv.num_edicts; j++)
    {
        ent = EDICT_NUM(j);

        if (ent->free)
            continue;

        if (ent->v.think == Num_SubRemove)
        {
            // possible gib - the following conditions will 99.9999% of the time satisfy gibs
            if (ent->v.flags == 0 && ent->v.frame == 0 && ent->v.movetype == MOVETYPE_BOUNCE &&
                    ent->v.solid == SOLID_NOT)
                ent->v.think = Num_SubNull;

            //Con_Printf ("Possible Gib\n");
        }
    }
}


/*
PR_EdictWalk

Does a walk through all the edicts in the map, so we can manipulate them directly in the
engine rather than having to modify QC - helps retain mod compatibility
*/
void PR_EdictWalk (void)
{
    edict_t *ent;
    int i;

    for (i=0 ; i<sv.num_edicts ; i++)
    {
        ent = EDICT_NUM(i);

        if (ent->free)
            continue;

        // do what we want in here
    }
}


/*
PR_SetModel

A Modified version of the QC PF_setmodel function which can be used for changing an edicts
model directly from the engine
*/
void PR_SetModel (edict_t *e, char *m)
{
    char	**check;
    model_t	*mod;
    int		i;

    // check to see if model was properly precached
    for (i=0, check = sv.model_precache ; *check ; i++, check++)
        if (!strcmp(*check, m))
            break;

    if (!*check)
        PR_RunError ("no precache: %s\n", m);

    e->v.model = m - pr_strings;
    e->v.modelindex = i;

    mod = sv.models[(int)e->v.modelindex];

    if (mod)
        SetMinMaxSize (e, mod->mins, mod->maxs, true);
    else SetMinMaxSize (e, vec3_origin, vec3_origin, true);
}

