/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// sbar.c -- status bar code

// MH - New HUD Drawing routines

#include "quakedef.h"

extern int oldtexture;

cvar_t	cl_fpscount = {"cl_fpscount", "0", true};
cvar_t	sbar_drawtype = {"sbar_drawtype", "0", true};

int			sb_updates;		// if >= vid.numpages, no update needed

// so we can update the fps counter per second rather than per frame
float second_count;
int old_fpscount;
float sbcf_secondcount;

#define STAT_MINUS		10	// num frame for '-' stats digit
qpic_t		*sb_nums[2][11];
qpic_t		*sb_colon, *sb_slash;
qpic_t		*sb_ibar;
qpic_t		*sb_sbar;
qpic_t		*sb_scorebar;

qpic_t      *sb_weapons[7][8];   // 0 is active, 1 is owned, 2-5 are flashes
qpic_t      *sb_ammo[4];
qpic_t		*sb_sigil[4];
qpic_t		*sb_armor[3];
qpic_t		*sb_items[32];

qpic_t	*sb_faces[7][2];		// 0 is gibbed, 1 is dead, 2-6 are alive
// 0 is static, 1 is temporary animation
qpic_t	*sb_face_invis;
qpic_t	*sb_face_quad;
qpic_t	*sb_face_invuln;
qpic_t	*sb_face_invis_invuln;

qboolean	sb_showscores;

int			sb_lines;			// scan lines to draw
int			xhair_xpos;
int			xhair_ypos;

qpic_t      *rsb_invbar[2];
qpic_t      *rsb_weapons[5];
qpic_t      *rsb_items[2];
qpic_t      *rsb_ammo[3];
qpic_t      *rsb_teambord;		// PGM 01/19/97 - team color border

//MED 01/04/97 added two more weapons + 3 alternates for grenade launcher
qpic_t      *hsb_weapons[7][5];   // 0 is active, 1 is owned, 2-5 are flashes
//MED 01/04/97 added array to simplify weapon parsing
int         hipweapons[4] = {HIT_LASER_CANNON_BIT,HIT_MJOLNIR_BIT,4,HIT_PROXIMITY_GUN_BIT};
//MED 01/04/97 added hipnotic items array
qpic_t      *hsb_items[2];

void Sbar_DeathmatchOverlay (void);
void M_DrawPic (int x, int y, qpic_t *pic);
int xhairtexture;
extern cvar_t viewsize;

extern byte xhairpic[16][16];
extern byte xhairpic_hires[16][16];

int sb_colorflash;

void R_InitXHairTexture (void)
{
    int x, y;
    byte	data[16][16][4];

    xhairtexture = texture_extension_number++;
    GL_Bind (xhairtexture);

    for (x=0 ; x<16 ; x++)
    {
        for (y=0 ; y<16 ; y++)
        {
            data[y][x][0] =
                data[y][x][1] =
                    data[y][x][2] = 255;
            if (vid.width == 640)
                data[y][x][3] = xhairpic[x][y] * 255;
            else data[y][x][3] = xhairpic_hires[x][y] * 255;
        }
    }

    glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 16, 16, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


/*
===============
Sbar_ShowScores

Tab key down
===============
*/
void Sbar_ShowScores (void)
{
    sbar_drawtype.value = sbar_drawtype.value + 1;

    if (sbar_drawtype.value > 1)
        sbar_drawtype.value = 0;

    Cvar_SetValue ("sbar_drawtype", sbar_drawtype.value);
}


// this isn't actually used in mhquake,but the function is here to prevent unnecessary
// con_printfs every coupla seconds - which would cause a major speed hit.
void Sbar_DontShowScores (void)
{
}


// viewsize now controls how much of the status bar is drawn
// 100: all of it
// 110: no weapon icons (top left)
// 120: none of it
void SCR_SizeUp_f (void)
{
    Cvar_SetValue ("viewsize", viewsize.value + 10);
    if (viewsize.value > 120) Cvar_SetValue ("viewsize", 120);
}


void SCR_SizeDown_f (void)
{
    Cvar_SetValue ("viewsize", viewsize.value - 10);
    if (viewsize.value < 100) Cvar_SetValue ("viewsize", 100);
}


/*
===============
Sbar_Init
===============
*/
void Sbar_Init (void)
{
    int		i;
    float	xhair_xcoord_mody, xhair_ycoord_mody;

    // set crosshair coordinates as accurate as possible for all resolutions, otherwise it
    // will drift slightly towards the bottom right corner as you increase the resolution
    // glquake draws relative to 640*480 so use that as a base
    xhair_xcoord_mody = 8.0 / 640.0 * (float) vid.width;
    xhair_ycoord_mody = 8.0 / 480.0 * (float) vid.height;

    // subtract 1 cos on screen coordinates begin at 0,0, not 1,1
    xhair_xpos = (vid.width / 2) - (int) xhair_xcoord_mody - 1;
    xhair_ypos = (vid.height / 2) - (int) xhair_ycoord_mody - 1;

    //Con_Printf ("Crosshair Position: x=%i y=%i\n", xhair_xpos, xhair_ypos);

    for (i=0 ; i<10 ; i++)
    {
        sb_nums[0][i] = Draw_PicFromWad (va("num_%i",i));
        sb_nums[1][i] = Draw_PicFromWad (va("anum_%i",i));
    }

    sb_nums[0][10] = Draw_PicFromWad ("num_minus");
    sb_nums[1][10] = Draw_PicFromWad ("anum_minus");

    sb_colon = Draw_PicFromWad ("num_colon");
    sb_slash = Draw_PicFromWad ("num_slash");

    sb_weapons[0][0] = Draw_PicFromWad ("inv_shotgun");
    sb_weapons[0][1] = Draw_PicFromWad ("inv_sshotgun");
    sb_weapons[0][2] = Draw_PicFromWad ("inv_nailgun");
    sb_weapons[0][3] = Draw_PicFromWad ("inv_snailgun");
    sb_weapons[0][4] = Draw_PicFromWad ("inv_rlaunch");
    sb_weapons[0][5] = Draw_PicFromWad ("inv_srlaunch");
    sb_weapons[0][6] = Draw_PicFromWad ("inv_lightng");

    sb_weapons[1][0] = Draw_PicFromWad ("inv2_shotgun");
    sb_weapons[1][1] = Draw_PicFromWad ("inv2_sshotgun");
    sb_weapons[1][2] = Draw_PicFromWad ("inv2_nailgun");
    sb_weapons[1][3] = Draw_PicFromWad ("inv2_snailgun");
    sb_weapons[1][4] = Draw_PicFromWad ("inv2_rlaunch");
    sb_weapons[1][5] = Draw_PicFromWad ("inv2_srlaunch");
    sb_weapons[1][6] = Draw_PicFromWad ("inv2_lightng");

    for (i=0 ; i<5 ; i++)
    {
        sb_weapons[2+i][0] = Draw_PicFromWad (va("inva%i_shotgun",i+1));
        sb_weapons[2+i][1] = Draw_PicFromWad (va("inva%i_sshotgun",i+1));
        sb_weapons[2+i][2] = Draw_PicFromWad (va("inva%i_nailgun",i+1));
        sb_weapons[2+i][3] = Draw_PicFromWad (va("inva%i_snailgun",i+1));
        sb_weapons[2+i][4] = Draw_PicFromWad (va("inva%i_rlaunch",i+1));
        sb_weapons[2+i][5] = Draw_PicFromWad (va("inva%i_srlaunch",i+1));
        sb_weapons[2+i][6] = Draw_PicFromWad (va("inva%i_lightng",i+1));
    }

    sb_ammo[0] = Draw_PicFromWad ("sb_shells");
    sb_ammo[1] = Draw_PicFromWad ("sb_nails");
    sb_ammo[2] = Draw_PicFromWad ("sb_rocket");
    sb_ammo[3] = Draw_PicFromWad ("sb_cells");

    sb_armor[0] = Draw_PicFromWad ("sb_armor1");
    sb_armor[1] = Draw_PicFromWad ("sb_armor2");
    sb_armor[2] = Draw_PicFromWad ("sb_armor3");

    sb_items[0] = Draw_PicFromWad ("sb_key1");
    sb_items[1] = Draw_PicFromWad ("sb_key2");
    sb_items[2] = Draw_PicFromWad ("sb_invis");
    sb_items[3] = Draw_PicFromWad ("sb_invuln");
    sb_items[4] = Draw_PicFromWad ("sb_suit");
    sb_items[5] = Draw_PicFromWad ("sb_quad");

    sb_sigil[0] = Draw_PicFromWad ("sb_sigil1");
    sb_sigil[1] = Draw_PicFromWad ("sb_sigil2");
    sb_sigil[2] = Draw_PicFromWad ("sb_sigil3");
    sb_sigil[3] = Draw_PicFromWad ("sb_sigil4");

    sb_faces[4][0] = Draw_PicFromWad ("face1");
    sb_faces[4][1] = Draw_PicFromWad ("face_p1");
    sb_faces[3][0] = Draw_PicFromWad ("face2");
    sb_faces[3][1] = Draw_PicFromWad ("face_p2");
    sb_faces[2][0] = Draw_PicFromWad ("face3");
    sb_faces[2][1] = Draw_PicFromWad ("face_p3");
    sb_faces[1][0] = Draw_PicFromWad ("face4");
    sb_faces[1][1] = Draw_PicFromWad ("face_p4");
    sb_faces[0][0] = Draw_PicFromWad ("face5");
    sb_faces[0][1] = Draw_PicFromWad ("face_p5");

    sb_face_invis = Draw_PicFromWad ("face_invis");
    sb_face_invuln = Draw_PicFromWad ("face_invul2");
    sb_face_invis_invuln = Draw_PicFromWad ("face_inv2");
    sb_face_quad = Draw_PicFromWad ("face_quad");

    Cmd_AddCommand ("+showscores", Sbar_ShowScores);
    Cmd_AddCommand ("-showscores", Sbar_DontShowScores);

    sb_sbar = Draw_PicFromWad ("sbar");
    sb_ibar = Draw_PicFromWad ("ibar");
    sb_scorebar = Draw_PicFromWad ("scorebar");

//MED 01/04/97 added new hipnotic weapons
    if (hipnotic)
    {
        hsb_weapons[0][0] = Draw_PicFromWad ("inv_laser");
        hsb_weapons[0][1] = Draw_PicFromWad ("inv_mjolnir");
        hsb_weapons[0][2] = Draw_PicFromWad ("inv_gren_prox");
        hsb_weapons[0][3] = Draw_PicFromWad ("inv_prox_gren");
        hsb_weapons[0][4] = Draw_PicFromWad ("inv_prox");

        hsb_weapons[1][0] = Draw_PicFromWad ("inv2_laser");
        hsb_weapons[1][1] = Draw_PicFromWad ("inv2_mjolnir");
        hsb_weapons[1][2] = Draw_PicFromWad ("inv2_gren_prox");
        hsb_weapons[1][3] = Draw_PicFromWad ("inv2_prox_gren");
        hsb_weapons[1][4] = Draw_PicFromWad ("inv2_prox");

        for (i=0 ; i<5 ; i++)
        {
            hsb_weapons[2+i][0] = Draw_PicFromWad (va("inva%i_laser",i+1));
            hsb_weapons[2+i][1] = Draw_PicFromWad (va("inva%i_mjolnir",i+1));
            hsb_weapons[2+i][2] = Draw_PicFromWad (va("inva%i_gren_prox",i+1));
            hsb_weapons[2+i][3] = Draw_PicFromWad (va("inva%i_prox_gren",i+1));
            hsb_weapons[2+i][4] = Draw_PicFromWad (va("inva%i_prox",i+1));
        }

        hsb_items[0] = Draw_PicFromWad ("sb_wsuit");
        hsb_items[1] = Draw_PicFromWad ("sb_eshld");
    }

    if (rogue)
    {
        rsb_invbar[0] = Draw_PicFromWad ("r_invbar1");
        rsb_invbar[1] = Draw_PicFromWad ("r_invbar2");

        rsb_weapons[0] = Draw_PicFromWad ("r_lava");
        rsb_weapons[1] = Draw_PicFromWad ("r_superlava");
        rsb_weapons[2] = Draw_PicFromWad ("r_gren");
        rsb_weapons[3] = Draw_PicFromWad ("r_multirock");
        rsb_weapons[4] = Draw_PicFromWad ("r_plasma");

        rsb_items[0] = Draw_PicFromWad ("r_shield1");
        rsb_items[1] = Draw_PicFromWad ("r_agrav1");

// PGM 01/19/97 - team color border
        rsb_teambord = Draw_PicFromWad ("r_teambord");
// PGM 01/19/97 - team color border

        rsb_ammo[0] = Draw_PicFromWad ("r_ammolava");
        rsb_ammo[1] = Draw_PicFromWad ("r_ammomulti");
        rsb_ammo[2] = Draw_PicFromWad ("r_ammoplasma");
    }

    Cvar_RegisterVariable (&cl_fpscount);
    Cvar_RegisterVariable (&sbar_drawtype);
    Cmd_AddCommand ("sizeup", SCR_SizeUp_f);
    Cmd_AddCommand ("sizedown", SCR_SizeDown_f);
    R_InitXHairTexture ();

    sb_colorflash = 0;
}


// draw the crosshair from the sbar rather than as normal... this way, we can ignore it
// properly if we have the axe
void SCR_Draw_XHair (int x, int y)
{
    glEnable (GL_BLEND);
    GL_Bind (xhairtexture);

    glBegin (GL_QUADS);

    glTexCoord2f (0, 0);
    glVertex2f (x, y);

    glTexCoord2f (1, 0);
    glVertex2f (x+16, y);

    glTexCoord2f (1, 1);
    glVertex2f (x+16, y+16);

    glTexCoord2f (0, 1);
    glVertex2f (x, y+16);

    glEnd ();
    glDisable (GL_BLEND);
}


//=============================================================================

/*
=============
Sbar_itoa
=============
*/
int Sbar_itoa (int num, char *buf)
{
    char	*str;
    int		pow10;
    int		dig;

    str = buf;

    if (num < 0)
    {
        *str++ = '-';
        num = -num;
    }

    for (pow10 = 10 ; num >= pow10 ; pow10 *= 10)
        ;

    do
    {
        pow10 /= 10;
        dig = num/pow10;
        *str++ = '0'+dig;
        num -= dig*pow10;
    }
    while (pow10 != 1);

    *str = 0;

    return str-buf;
}


/*
=============
Sbar_DrawNum
=============
*/
void Sbar_DrawNum (int x, int y, int num, int digits, int color)
{
    char			str[12];
    char			*ptr;
    int				l, frame;

    l = Sbar_itoa (num, str);
    ptr = str;

    if (l > digits)
        ptr += (l-digits);

    if (l < digits)
        x += (digits-l)*24;

    while (*ptr)
    {
        if (*ptr == '-')
            frame = STAT_MINUS;
        else
            frame = *ptr -'0';

        // use this rather than Sbar_DrawTransPic so we can draw relative to the screen
        // size, rather than relative to a status bar which no longer exists...
        Draw_TransPic (x, y, sb_nums[color && sb_colorflash][frame]);

        x += 24;
        ptr++;
    }
}

//=============================================================================

int		fragsort[MAX_SCOREBOARD];

char	scoreboardtext[MAX_SCOREBOARD][20];
int		scoreboardtop[MAX_SCOREBOARD];
int		scoreboardbottom[MAX_SCOREBOARD];
int		scoreboardcount[MAX_SCOREBOARD];
int		scoreboardlines;

/*
===============
Sbar_SortFrags
===============
*/
void Sbar_SortFrags (void)
{
    int		i, j, k;

    // sort by frags
    scoreboardlines = 0;
    for (i=0 ; i<cl.maxclients ; i++)
    {
        if (cl.scores[i].name[0])
        {
            fragsort[scoreboardlines] = i;
            scoreboardlines++;
        }
    }

    for (i=0 ; i<scoreboardlines ; i++)
        for (j=0 ; j<scoreboardlines-1-i ; j++)
            if (cl.scores[fragsort[j]].frags < cl.scores[fragsort[j+1]].frags)
            {
                k = fragsort[j];
                fragsort[j] = fragsort[j+1];
                fragsort[j+1] = k;
            }
}

/*
===============
Sbar_SoloScoreboard
===============
*/
void Sbar_SoloScoreboard (void)
{
    char	str[80];
    int		minutes, seconds, tens, units;
    int		l;

    sprintf (str,"Monsters:%3i /%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
    Draw_String ((vid.width - 320)/2+8, vid.height - 20, str);

    sprintf (str,"Secrets :%3i /%3i", cl.stats[STAT_SECRETS], cl.stats[STAT_TOTALSECRETS]);
    Draw_String ((vid.width - 320)/2+8, vid.height - 12, str);

    // time
    minutes = cl.time / 60;
    seconds = cl.time - 60*minutes;
    tens = seconds / 10;
    units = seconds - 10*tens;
    sprintf (str,"Time :%3i:%i%i", minutes, tens, units);
    Draw_String ((vid.width - 320)/2+184, vid.height - 20, str);

    // draw level name
    l = strlen (cl.levelname);
    Draw_String ((vid.width - 320)/2+232 - l*4, vid.height - 12, cl.levelname);
}

/*
===============
Sbar_DrawScoreboard
===============
*/
void Sbar_DrawScoreboard (void)
{
    // no scoreboard in deathmatch - cos it gives monsters and such, which is silly...
    if (cl.gametype == GAME_DEATHMATCH)
        return;

    Sbar_SoloScoreboard ();
}

//=============================================================================

/*
===============
Sbar_Draw
===============
*/
void Sbar_Draw (void)
{
    int flashon;
    int i;
    int anim, f;
    char num[6];
    int ammo_count;
    int fps;

    // always update these regardless of whether or not we're using them in the current frame,
    // so we always have an accurate value for them.
    second_count = second_count + frametime;
    sbcf_secondcount = sbcf_secondcount + frametime;

    // only update 10 times per frame
    if (second_count >= 0.1)
    {
        fps = 1 / frametime;
        old_fpscount = fps;
        second_count = 0;
    }
    else fps = old_fpscount;

    if (fps > 999) fps = 999;	// just in case...

    // ticks of 0.25 seconds
    if (sbcf_secondcount > 0.25)
    {
        sbcf_secondcount = 0;
        sb_colorflash = !sb_colorflash;
    }

    // most qc cutscenes set viewsize to 120 for no status bar during the cutscene - so do the
    // save here...
    if (viewsize.value > 110) return;

    // don't draw if a fullscreen console
    if (scr_con_current == vid.height) return;

    // if in deathmatch, draw the current fragtable (whether dead or not)
    if (cl.gametype == GAME_DEATHMATCH) Sbar_DeathmatchOverlay ();

    // if we've just died, show the scoreboard (don't draw the rest of the HUD)
    if (cl.stats[STAT_HEALTH] <= 0)
    {
        Sbar_DrawScoreboard ();
        return;
    }

    // if we explicitly wanna see the scoreboard, show it
    if (sbar_drawtype.value == 1) Sbar_DrawScoreboard ();

    // armour icon - use the same armour icons for normal quake, rogue and hipnotic
    if (cl.items & IT_INVULNERABILITY)
    {
        Draw_Pic (6 , vid.height - 60, draw_disc);
        Sbar_DrawNum (36, vid.height - 60, 666, 3, 1);
    }
    else if (cl.stats[STAT_ARMOR] > 0 && rogue)
    {
        // only draw the armor icon if there is an armor value!
        if (cl.items & RIT_ARMOR3) Draw_Pic (6 , vid.height - 60, sb_armor[2]);
        else if (cl.items & RIT_ARMOR2) Draw_Pic (6 , vid.height - 60, sb_armor[1]);
        else Draw_Pic (6 , vid.height - 60, sb_armor[0]);

        Sbar_DrawNum (36, vid.height - 60, cl.stats[STAT_ARMOR], 3, cl.stats[STAT_ARMOR] <= 25);
    }
    else if (cl.stats[STAT_ARMOR] > 0)
    {
        // only draw the armor icon if there is an armor value!
        if (cl.items & IT_ARMOR3) Draw_Pic (6 , vid.height - 60, sb_armor[2]);
        else if (cl.items & IT_ARMOR2) Draw_Pic (6 , vid.height - 60, sb_armor[1]);
        else Draw_Pic (6 , vid.height - 60, sb_armor[0]);

        Sbar_DrawNum (36, vid.height - 60, cl.stats[STAT_ARMOR], 3, cl.stats[STAT_ARMOR] <= 25);
    }

    // health icon - always drawn - this version does not do team colour drawing
    // get the right face for the current powerup, if applicable
    if ((cl.items & (IT_INVISIBILITY | IT_INVULNERABILITY)) == (IT_INVISIBILITY | IT_INVULNERABILITY))
        Draw_Pic (6, vid.height - 30, sb_face_invis_invuln);
    else if (cl.items & IT_QUAD)
        Draw_Pic (6, vid.height - 30, sb_face_quad);
    else if (cl.items & IT_INVISIBILITY)
        Draw_Pic (6, vid.height - 30, sb_face_invis);
    else if (cl.items & IT_INVULNERABILITY)
        Draw_Pic (6, vid.height - 30, sb_face_invuln);
    else
    {
        // find the correct animation
        if (cl.stats[STAT_HEALTH] >= 100)
            f = 4;
        else f = cl.stats[STAT_HEALTH] / 20;

        if (cl.time <= cl.faceanimtime)
            anim = 1;
        else anim = 0;

        Draw_Pic (6, vid.height - 30, sb_faces[f][anim]);
    }

    Sbar_DrawNum (36, vid.height - 30, cl.stats[STAT_HEALTH], 3, cl.stats[STAT_HEALTH] <= 25);

    // ammo icon
    // ammo - added rogue ammo icons - hipnotic is unchanged
    if (rogue)
    {
        if (cl.items & RIT_SHELLS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[0]);
        else if (cl.items & RIT_NAILS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[1]);
        else if (cl.items & RIT_ROCKETS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[2]);
        else if (cl.items & RIT_CELLS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[3]);
        else if (cl.items & RIT_LAVA_NAILS) Draw_Pic (vid.width - 30, vid.height - 30, rsb_ammo[0]);
        else if (cl.items & RIT_PLASMA_AMMO) Draw_Pic (vid.width - 30, vid.height - 30, rsb_ammo[1]);
        else if (cl.items & RIT_MULTI_ROCKETS) Draw_Pic (vid.width - 30, vid.height - 30, rsb_ammo[2]);
    }
    else
    {
        if (cl.items & IT_SHELLS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[0]);
        else if (cl.items & IT_NAILS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[1]);
        else if (cl.items & IT_ROCKETS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[2]);
        else if (cl.items & IT_CELLS) Draw_Pic (vid.width - 30, vid.height - 30, sb_ammo[3]);
    }

    // the stuff in here is only drawn if we don't have the axe
    if (cl.stats[STAT_AMMO] != 0)
    {
        // ammo count on the status bar
        Sbar_DrawNum (vid.width - 108, vid.height - 30, cl.stats[STAT_AMMO], 3, cl.stats[STAT_AMMO] <= 10);

        SCR_Draw_XHair (xhair_xpos, xhair_ypos);
    }

    // fps counter
    if (cl_fpscount.value)
    {
        // fixed long standing bug here - was %2i in the sprintf
        Draw_String (vid.width - 70, 6, "FPS:   ");
        sprintf (num, "%3i", fps);

        if (num[0] != ' ') Draw_Character (vid.width - 30, 6, 18 + num[0] - '0');
        if (num[1] != ' ') Draw_Character (vid.width - 22, 6, 18 + num[1] - '0');
        if (num[2] != ' ') Draw_Character (vid.width - 14, 6, 18 + num[2] - '0');
    }

    // Keys are identical - no need to change
    if (cl.items & IT_KEY1) Draw_Pic (vid.width - 44, vid.height - 50, sb_items[0]);
    if (cl.items & IT_KEY2) Draw_Pic (vid.width - 22, vid.height - 50, sb_items[1]);

    // Sigils
    if (!rogue && !hipnotic)
    {
        // don't draw sigils in rogue or hipnotic
        if (cl.items & IT_SIGIL1) Draw_Pic (vid.width - 64, vid.height - 50, sb_sigil[0]);
        if (cl.items & IT_SIGIL2) Draw_Pic (vid.width - 78, vid.height - 50, sb_sigil[1]);
        if (cl.items & IT_SIGIL3) Draw_Pic (vid.width - 92, vid.height - 50, sb_sigil[2]);
        if (cl.items & IT_SIGIL4) Draw_Pic (vid.width - 106, vid.height - 50, sb_sigil[3]);
    }

    if (viewsize.value > 100) return;

    // weaponses
    if (rogue)
    {
        for (i=0 ; i<7 ; i++)
        {
            // this is horrible stuff.
            if ((i == 2) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_LAVA_NAILGUN) && (cl.items & RIT_LAVA_NAILGUN))
                Draw_Pic (30, 46, rsb_weapons[0]);
            else if ((i == 3) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_LAVA_SUPER_NAILGUN) && (cl.items & RIT_LAVA_SUPER_NAILGUN))
                Draw_Pic (30, 66, rsb_weapons[1]);
            else if ((i == 4) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_MULTI_GRENADE) && (cl.items & RIT_MULTI_GRENADE))
                Draw_Pic (30, 86, rsb_weapons[2]);
            else if ((i == 5) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_MULTI_ROCKET) && (cl.items & RIT_MULTI_ROCKET))
                Draw_Pic (30, 106, rsb_weapons[3]);
            else if ((i == 6) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_PLASMA_GUN) && (cl.items & RIT_PLASMA_GUN))
                Draw_Pic (6, 126, rsb_weapons[4]);
            else if (cl.items & (IT_SHOTGUN<<i) )
            {
                flashon = (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i));
                if (i == 6) Draw_Pic (6, 6+(i*20), sb_weapons[flashon][i]);
                else Draw_Pic (30, 6+(i*20), sb_weapons[flashon][i]);
            }
            else
                continue;

            ammo_count = i/2;  // because it's an int, it will truncate instead of round up
            sprintf (num, "%3i",cl.stats[STAT_SHELLS+ammo_count]);

            if (num[0] != ' ') Draw_Character ( 60, 8+(i*20), 176 + num[0] - '0');
            if (num[1] != ' ') Draw_Character ( 68, 8+(i*20), 176 + num[1] - '0');
            if (num[2] != ' ') Draw_Character ( 76, 8+(i*20), 176 + num[2] - '0');
        }
    }
    else
    {
        for (i=0 ; i<7 ; i++)
        {
            if (cl.items & (IT_SHOTGUN<<i) )
            {
                flashon = (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i));
                if (i == 6 && !hipnotic) Draw_Pic (6, 6+(i*20), sb_weapons[flashon][i]);
                else if (i == 6 && hipnotic) Draw_Pic (22, 6+(i*20), sb_weapons[flashon][i]);
                else Draw_Pic (30, 6+(i*20), sb_weapons[flashon][i]);

                // ammo counts
                ammo_count = i/2;  // because it's an int, it will truncate instead of round up
                sprintf (num, "%3i",cl.stats[STAT_SHELLS+ammo_count]);

                if (num[0] != ' ') Draw_Character ( 60, 8+(i*20), 176 + num[0] - '0');
                if (num[1] != ' ') Draw_Character ( 68, 8+(i*20), 176 + num[1] - '0');
                if (num[2] != ' ') Draw_Character ( 76, 8+(i*20), 176 + num[2] - '0');
            }
        }
    }

    // hipnotic extra weapons
    if (hipnotic)
    {
        // proxy gun - alt for weapon 6 - this is horrible
        if (cl.items & HIT_PROXIMITY_GUN)
        {
            if (cl.stats[STAT_ACTIVEWEAPON] == HIT_PROXIMITY_GUN)
            {
                Draw_Pic (30, 86, hsb_weapons[1][2]);
                sprintf (num, "%3i",cl.stats[STAT_ROCKETS]);
                if (num[0] != ' ') Draw_Character ( 60, 88, 176 + num[0] - '0');
                if (num[1] != ' ') Draw_Character ( 68, 88, 176 + num[1] - '0');
                if (num[2] != ' ') Draw_Character ( 76, 88, 176 + num[2] - '0');
            }
            else if (!(cl.items & IT_GRENADE_LAUNCHER))
            {
                Draw_Pic (30, 86, hsb_weapons[0][2]);
                sprintf (num, "%3i",cl.stats[STAT_ROCKETS]);
                if (num[0] != ' ') Draw_Character ( 60, 88, 176 + num[0] - '0');
                if (num[1] != ' ') Draw_Character ( 68, 88, 176 + num[1] - '0');
                if (num[2] != ' ') Draw_Character ( 76, 88, 176 + num[2] - '0');
            }
        }

        // mjolnir - weapon 0
        if (cl.items & HIT_MJOLNIR)
        {
            flashon = (cl.stats[STAT_ACTIVEWEAPON] == HIT_MJOLNIR);
            Draw_Pic (30, 166, hsb_weapons[flashon][1]);
            sprintf (num, "%3i",cl.stats[STAT_CELLS]);
            if (num[0] != ' ') Draw_Character ( 60, 168, 176 + num[0] - '0');
            if (num[1] != ' ') Draw_Character ( 68, 168, 176 + num[1] - '0');
            if (num[2] != ' ') Draw_Character ( 76, 168, 176 + num[2] - '0');
        }

        // laser cannon - weapon 9
        if (cl.items & HIT_LASER_CANNON)
        {
            flashon = (cl.stats[STAT_ACTIVEWEAPON] == HIT_LASER_CANNON);
            Draw_Pic (30, 146, hsb_weapons[flashon][0]);
            sprintf (num, "%3i",cl.stats[STAT_CELLS]);
            if (num[0] != ' ') Draw_Character ( 60, 148, 176 + num[0] - '0');
            if (num[1] != ' ') Draw_Character ( 68, 148, 176 + num[1] - '0');
            if (num[2] != ' ') Draw_Character ( 76, 148, 176 + num[2] - '0');
        }
    }
}


//=============================================================================

/*
==================
Sbar_IntermissionNumber

==================
*/
void Sbar_IntermissionNumber (int x, int y, int num, int digits, int color)
{
    char			str[12];
    char			*ptr;
    int				l, frame;

    l = Sbar_itoa (num, str);
    ptr = str;
    if (l > digits)
        ptr += (l-digits);
    if (l < digits)
        x += (digits-l)*24;

    while (*ptr)
    {
        if (*ptr == '-')
            frame = STAT_MINUS;
        else
            frame = *ptr -'0';

        Draw_TransPic (x,y,sb_nums[color][frame]);
        x += 24;
        ptr++;
    }
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_DeathmatchOverlay (void)
{
    int				i, k, l;
    int				x, y, f;
    char			num[12];
    scoreboard_t	*s;

    scr_copyeverything = 1;
    scr_fullupdate = 0;

    //pic = Draw_CachePic ("gfx/ranking.lmp");
    //M_DrawPic ((320-pic->width)/2, 8, pic);

    // scores
    Sbar_SortFrags ();

    // draw the text
    l = scoreboardlines;

    // draw the deathmatch overlay in the bottom left corner just over the HUD
    // this makes the mini-overlay redundant
    //x = 80 + ((vid.width - 320)>>1);
    //y = 40;
    x = 6;
    y = vid.height - 68 - (8 * l);

    for (i=0 ; i<l ; i++)
    {
        k = fragsort[i];
        s = &cl.scores[k];

        if (!s->name[0])
            continue;

        // draw number
        f = s->frags;
        sprintf (num, "%3i",f);

        Draw_Character ( x , y, num[0]);
        Draw_Character ( x+8 , y, num[1]);
        Draw_Character ( x+16 , y, num[2]);

        // no idea what this is, i think it's colouring differently to indicate the player
        //if (k == cl.viewentity - 1)
        if (!strcmp (s->name, cl_name.string))
            Draw_Character ( x + 32, y, 12);

        // draw name
        Draw_String (x+40, y, s->name);

        y += 8;
    }
}


/*
==================
Sbar_IntermissionOverlay

==================
*/
void Sbar_IntermissionOverlay (void)
{
    qpic_t	*pic;
    int		dig;
    int		num;

    scr_copyeverything = 1;
    scr_fullupdate = 0;

    if (cl.gametype == GAME_DEATHMATCH)
    {
        Sbar_DeathmatchOverlay ();
        return;
    }

    pic = Draw_CachePic ("gfx/complete.lmp");
    Draw_Pic (64, 24, pic);

    pic = Draw_CachePic ("gfx/inter.lmp");
    Draw_TransPic (0, 56, pic);

// time
    dig = cl.completed_time/60;
    Sbar_IntermissionNumber (160, 64, dig, 3, 0);
    num = cl.completed_time - dig*60;
    Draw_TransPic (234,64,sb_colon);
    Draw_TransPic (246,64,sb_nums[0][num/10]);
    Draw_TransPic (266,64,sb_nums[0][num%10]);

    Sbar_IntermissionNumber (160, 104, cl.stats[STAT_SECRETS], 3, 0);
    Draw_TransPic (232,104,sb_slash);
    Sbar_IntermissionNumber (240, 104, cl.stats[STAT_TOTALSECRETS], 3, 0);

    Sbar_IntermissionNumber (160, 144, cl.stats[STAT_MONSTERS], 3, 0);
    Draw_TransPic (232,144,sb_slash);
    Sbar_IntermissionNumber (240, 144, cl.stats[STAT_TOTALMONSTERS], 3, 0);

}


/*
==================
Sbar_FinaleOverlay

==================
*/
void Sbar_FinaleOverlay (void)
{
    qpic_t	*pic;

    scr_copyeverything = 1;

    pic = Draw_CachePic ("gfx/finale.lmp");
    Draw_TransPic ( (vid.width-pic->width)/2, 16, pic);
}
